from boa.blockchain.vm.Neo.Runtime import GetTrigger, CheckWitness
from boa.blockchain.vm.Neo.TriggerType import Application, Verification
from boa.blockchain.vm.Neo.Runtime import Notify


OWNER = b'\xaf\x12\xa8h{\x14\x94\x8b\xc4\xa0\x08\x12\x8aU\nci[\xc1\xa5'


def Main(a, b, c, d):

    """

    :param a:
    :param b:
    :param c:
    :param d:
    :return:
    """
    print("Executing Four Param Verification code")
    trigger = GetTrigger()

    if trigger == Verification():

        print("Running verification")
        is_owner = CheckWitness(OWNER)
        Notify(is_owner)
        return is_owner

    print("runnig app routine")

    result = a + b - c + d

    Notify(result)

    return result




from boa.code.builtins import list, range
from boa.blockchain.vm.Neo.Runtime import Log, Notify


def Main():

    """

    :return:
    """
    mmm = range(2, 14)

    l = mmm[2]

    Notify(l)

    empty = list()

    b = get_thing()
    c = get_items_from_range(mmm, 7)

    k = range(10, 12)

    return l + b + c + k[0]


def get_thing():
    """

    :return:
    """
    return 7


def get_items_from_range(items, index):

    """

    :param items:
    :param index:
    :return:
    """
    return items[index]




class SmartContract():

    def Sha1(data):
        """

        """
        pass

    def Sha256(data):
        """

        """
        pass

    def Hash160(data):
        """

        """
        pass

    def Hash256(data):
        """

        """
        pass

    def VerifySignature(pubkey, signature):
        """

        :param signature:
        """
        pass




def MinerTransaction():
    """

    :return:
    """
    return b'\x00'


def IssueTransaction():
    """

    :return:
    """
    return b'\x01'


def ClaimTransaction():
    """

    :return:
    """
    return b'\x02'


def EnrollmentTransaction():
    """

    :return:
    """
    return b'\x20'


def VotingTransaction():
    """

    :return:
    """
    return b'\x24'


def RegisterTransaction():
    """

    :return:
    """
    return b'\x40'


def ContractTransaction():
    """

    :return:
    """
    return b'\x80'


def AgencyTransaction():
    """

    :return:
    """
    return b'\xb0'


def PublishTransaction():
    """

    :return:
    """
    return b'\xd0'


def InvocationTransaction():
    """

    :return:
    """
    return b'\xd1'





def Main():

    """

    :return:
    """
    a = 3

    b = 2

    m = 12

    if not a == b:   # this currently works
        print("a not equal to b!!!")
        m = 21

    if a != b:

        print("numbers 2 and 3 are not equal")
        m = 82

    j = 'hello'
    k = 'hello'

    if j != k:

        print("string j is not equal to string k")

    else:

        print("string j is equal to string k")

    q = bytearray(b'\x10\x01\x80')
    q2 = bytearray(b'\x10\x10\x80')

    if q != q2:

        print("bytearrays m and m2 not equal")

    else:

        print("bytearrays m and m2 are equal")

    return m



from boa.blockchain.vm.Neo.Runtime import Notify


def Main():

    """

    :return:
    """
    items = [1, 3]

#    j = 0
#    for i in items:
#        j += i
#    i2 = [i+1 if i > 0 else i + 5 for i in items]

#    m = 0
    [Notify(i) for i in items]  # this is not working

    m = 3

    return m




class BigInteger(int):

    @staticmethod
    def FromBytes(data, signed=False):
        """
        Convert a bytearray into a BigInteger object

        :param data: a bytearray representing an integer
        :type data: bytearray

        :param signed: whether or not the bytearray is signed
        :type signed: bool

        :return: a BigInteger object
        :rtype: ``boa.blockchain.vm.BigInteger``

        """
        return BigInteger(int.from_bytes(data, 'little', signed=signed))

    def Equals(self, other):
        """
        Compare two BigInteger objects

        :param other: the BigInteger to compare this one with
        :type other: BigInteger

        :return: whether the two items are equal
        :rtype: bool

        """
        return super(BigInteger, self).__eq__(other)

    def ToByteArray(self, signed=True):

        """

        converts a big integer object into a bytearray

        :param signed: whether or not it should be signed
        :type signed: bool

        :return: a bytearray of representing the BigInteger
        :rtype: bytearray

        """

        if self < 0:
            try:
                return self.to_bytes(1 + ((self.bit_length() + 7) // 8), byteorder='little', signed=True)
            except Exception as e:
                print("coludnt convert negative number %s " % e)
                return False
        try:
            return self.to_bytes((self.bit_length() + 7) // 8, byteorder='little', signed=signed)
        except OverflowError:
            return self.to_bytes(1 + ((self.bit_length() + 7) // 8), byteorder='little', signed=signed)
        except Exception:
            print("COULD NOT CONVERT %s to byte array" % self)

    def __abs__(self, *args, **kwargs):  # real signature unknown
        return BigInteger(super(BigInteger, self).__abs__(*args, **kwargs))

    def __add__(self, *args, **kwargs):  # real signature unknown
        return BigInteger(super(BigInteger, self).__add__(*args, **kwargs))

    def __mod__(self, *args, **kwargs):  # real signature unknown
        return BigInteger(super(BigInteger, self).__mod__(*args, **kwargs))

    def __mul__(self, *args, **kwargs):  # real signature unknown
        return BigInteger(super(BigInteger, self).__mul__(*args, **kwargs))

    def __neg__(self, *args, **kwargs):  # real signature unknown
        return BigInteger(super(BigInteger, self).__neg__(*args, **kwargs))

    def __str__(self, *args, **kwargs):  # real signature unknown
        return super(BigInteger, self).__str__(*args, **kwargs)

    def __sub__(self, *args, **kwargs):  # real signature unknown
        return BigInteger(super(BigInteger, self).__sub__(*args, **kwargs))

    def __truediv__(self, *args, **kwargs):  # real signature unknown
        return BigInteger(int(super(BigInteger, self).__truediv__(*args, **kwargs)))


ZERO = BigInteger(0)
ONE = BigInteger(1)




def Main(a, b, c, d):

    """

    :param a:
    :param b:
    :param c:
    :param d:
    :return:
    """
    m = 0

    if a > b:

        if c > d:

            m = 3

        else:

            if b > c:

                return 8

            else:

                return 10

    else:

        if c > d:

            m = 1

        else:

            if b < c:

                return 11

            else:

                m = 22

    return m



from boa.blockchain.vm.Neo.Runtime import Log, Notify
from boa.code.builtins import list


def Main():

    """

    :return:
    """
    a = 1

    b = [1, 2, 34]

    b1 = b[0]
    b2 = b[1]
    c = add3(b1, b2)

    d = addd(a, c)
    print("will return")
    return a + c + d


def addd(a, b):

    """

    :param a:
    :param b:
    :return:
    """
    result = a + b
    print("addd")
    res2 = stuff()
    Log("add loggg")
    res3 = add3(result, res2)

    print("res 3 is...")
    Notify(res3)

    return result + stuff() + res2 + res3


def stuff():

    """

    :return:
    """
    a = 4

    b = 2

    items = [0, 1, 0, 9, 20, 4, 23]

    j = items[3]

    q = n_range(1, 10)

    return a + b + j + items[1] + q[4]


def add3(a, b):

    """

    :param a:
    :param b:
    :return:
    """
    print("add3")

    print("add 3 loggg")
    return a + b


def n_range(start, stop):
    """
    range(start, stop) -> list object

    Return an list that is a a sequence of integers from start (inclusive)
    to stop (exclusive).  range(i, j) produces i, i+1, i+2, ..., j-1.
    """

#    out = []

#    if start >= stop:
#        return out

    length = stop - start

    out = list(length=length)

    index = 0
    orig_start = start

    while start < stop:
        val = index + orig_start
        out[index] = val
        index = index + 1
        start = orig_start + index

    return out





def Main():

    """
    a sample smart contract demonstrating how to use lambdas in smart contracts

    :return: result of the smart contract execution
    :rtype: int

    """
    j = 9

    def q(x):
        """

        :param x:
        :return:
        """
        return x + 1

    m = q(j)

    return m



from boa.blockchain.vm.Neo.Runtime import Notify
from boa.blockchain.vm.Neo.Storage import Get,GetContext,Put
from boa.code.builtins import concat,list,range,take,substr


# this is the same thing as SerializationTest1, but using slice notation instead of substr

SERIALIZED_NAME = 'AWESOME'

def Main():


    # create an array
    stuff = ['a','def','ghi','jk','lmnopqr']

    # serialize it
    to_save = serialize_array(stuff)


    # now lets store it
    context = GetContext()
    Put(context, SERIALIZED_NAME, to_save)

    # now retrieve it from storage
    serialized = Get(context, SERIALIZED_NAME)

    # now rebuild the array
    re_constructed = deserialize_bytearray(serialized)


    # iterate over the array to show its items match ``stuff``
    for item in re_constructed:
        Notify(item)

    # return the reconstructed array
    return re_constructed


def deserialize_bytearray(data):

    # ok this is weird.  if you remove this print statement, it stops working :/

    # get length of length
    collection_length_length = data[0:1]

    # get length of collection
    collection_len = data[1:collection_length_length+1]

    # create a new collection
    new_collection = list(length=collection_len)

    # trim the length data
    offset = 1 + collection_length_length

    for i in range(0, collection_len):

        # get the data length length
        itemlen_len = data[offset:offset+1]

        # get the length of the data
        item_len = data[offset+1:offset+1+itemlen_len]

        # get the data
        item = data[offset+1+itemlen_len: offset+1+itemlen_len+item_len]

        # store it in collection
        new_collection[i] = item

        offset = offset + item_len + itemlen_len + 1



    return new_collection


def serialize_array(items):


    # serialize the length of the list
    itemlength = serialize_var_length_item(items)


    output = itemlength


    # now go through and append all your stuff
    for item in items:

        # get the variable length of the item
        # to be serialized
        itemlen = serialize_var_length_item(item)

        # add that indicator
        output = concat(output, itemlen)

        # now add the item
        output = concat(output, item)


    # return the stuff
    return output


def serialize_var_length_item(item):

    # get the length of your stuff
    stuff_len = len(item)

    # now we need to know how many bytes the length of the array
    # will take to store

    # this is one byte
    if stuff_len <= 255:
        byte_len = b'\x01'
    # two byte
    elif stuff_len <= 65535:
        byte_len = b'\x02'
    # hopefully 4 byte
    else:
        byte_len = b'\x04'


    out = concat(byte_len, stuff_len)

    return out


import os
from boa.code.module import Module


class Compiler():
    """
    The main compiler interface class.

    The following loads a python file, compiles it to the `.avm` format
    and saves it alongside the python file.
    
    .. code-block:: python

        from boa.compiler import Compiler
        Compiler.load_and_save('path/to/your/file.py')

        # return the compiler object for inspection
        compiler = Compiler.load('path/to/your/file.py')

        # retrieve the default module for inpection
        default_module = compiler.default

        # retreive the default/entry method for the smart contract
        entry_method = default_module.main
    """
    
    __instance = None
    
    modules = None

    def __init__(self):
        self.modules = []
        

    @staticmethod
    def instance():
        """
        Retrieve the current instance of the Compiler object, if it exists,
        or create one.

        :return: the singleton instance of the Compiler object
        """
        
        if not Compiler.__instance:
            Compiler.__instance = Compiler()
        return Compiler.__instance
        

    @property
    def default(self):
        """
        Retrieve the default or 'entry' module.

        :return: the default `boa.code.Module` object or None upon exception
        """
        
        try:
            return self.modules[0]
        except Exception as e:
            pass
        

    @staticmethod
    def write_file(data, path):
        """
        Save the output data to the file system at the specified path.

        :param data: a byte string of data to write to disk
        :param path: the path to write the file to
        """
        
        with open(path, 'wb+') as out_file:
            out_file.write(data)
        

    def write(self):
        """
        Write the default module to a byte string.

        :return: the compiled Python program as a byte string
        :rtype: bytes
        """
        
        module = self.default
        out_bytes = bytes(module.write())

        return out_bytes
        

    @staticmethod
    def load_and_save(path, output_path=None):
        """
        Call `load_and_save` to load a Python file to be compiled to the .avm format and save the result.
        By default, the resultant .avm file is saved along side the source file.

        :param path: The path of the Python file to compile
        :param output_path: Optional path to save the compiled `.avm` file
        :return: the instance of the compiler

        The following returns the compiler object for inspection

        .. code-block:: python

            from boa.compiler import Compiler
        
            Compiler.load_and_save('path/to/your/file.py')
        """

        compiler = Compiler.load(path)
        data = compiler.write()

        if output_path is None:   
            fullpath = os.path.realpath(path)
            path, filename = os.path.split(fullpath)
            newfilename = filename.replace('.py', '.avm')
            output_path = '%s/%s' % (path, newfilename)
        
        Compiler.write_file(data, output_path)

        return data


    @staticmethod
    def load(path):
        """
        Call `load` to load a Python file to be compiled but not to write to .avm

        :param path: the path of the Python file to compile
        :return: The instance of the compiler

        The following returns the compiler object for inspection.

        .. code-block:: python

            from boa.compiler import Compiler
        
            compiler = Compiler.load('path/to/your/file.py')
        """

        Compiler.__instance = None

        compiler = Compiler.instance()

        module = Module(path)
        compiler.modules.append(module)

        return compiler



from boa.blockchain.vm.Neo.Storage import Get, Put, Delete, GetContext
from boa.blockchain.vm.Neo.Runtime import Notify


def Main():

    """

    :return:
    """
    context = GetContext()

    print("hello")
    Notify(context)

    item_key = 'hello'
    item_val = 'hhhhhhh'
    Notify(item_val)
    Put(context, item_key, item_val)
    print("hhhh")
    a = 1

    out = Get(context, item_key)

    return out




from boa.blockchain.vm.Neo.Action import RegisterAction


Transfer = RegisterAction('transfer', 'from', 'to', 'amount')

Refund = RegisterAction('refund', 'to', 'amount')


def Main():

    """

    :return:
    """
    a = 2

    b = 5

    c = a + b

    Transfer(a, b, c)

    to = 'me'
    amount = 52

    Refund(to, amount)

    d = Second(a)

    return d + 3


def Second(a):

    """

    :param a:
    :return:
    """
    j = 'hello'
    b = 'goodbye'

    Transfer(a, j, b)

    return 2





CONST = 8
OTHERCONTS = 1232

WHAT = 5


def Main():

    """

    :return:
    """
    print("do main?")
    j = CONST

    b = bleh()

    return CONST + OTHERCONTS + WHAT + b + j


def bleh():
    """

    :return:
    """
    return 2 + WHAT



from boa.code.builtins import concat
from boa.blockchain.vm.Neo.Runtime import Notify


def Main(a, b):

    """

    :param a:
    :param b:
    :return:
    """
    c = concat(a, b)

    Notify(c)

    if c == 'hellogoodbye':

        return 3

    return 1





def Main():

    """

    :return:
    """
    a = 1

    b = 10

    c = 20

    d = add(a, b, 10)

    d2 = add(d, d, d)

    return d2


def add(a, b, c):

    """

    :param a:
    :param b:
    :param c:
    :return:
    """
    result = a + b + c

    return result




def Main(fibnumber):

    """

    :param fibnumber:
    :return:
    """
    fibresult = fib(fibnumber)

    return fibresult


def fib(n):

    """

    :param n:
    :return:
    """
    if n == 1 or n == 2:
        return 1

    n1 = n - 1
    n2 = n - 2

    fibr1 = fib(n1)
    fibr2 = fib(n2)

    res = fibr1 + fibr2

    return res



from boa.blockchain.vm.Neo.Blockchain import GetBlock
from boa.blockchain.vm.Neo.Runtime import Notify, Log


def Main():

    """

    :return:
    """
    block_height = 123234
    block = GetBlock(block_height)

    print("hello")
    Notify(block)

    a = 1

    return block_height




def Main(a, b, c, d):
    """
    :param a: an integer
    :param b: an integer
    :param c: an integer
    :param d: an integer
    :return: (a * 2) + (b + 1) + (c / 2) + (d - 1)
    :rtype: int
    """
    
    a2 = a * 2

    b2 = b + 1

    c2 = c / 2

    d2 = d - 1

    return a2 + b2 + c2 + d2



from boa.code import pyop

from byteplay3 import Label

from boa.blockchain.vm import VMOp
from boa.blockchain.vm.BigInteger import BigInteger

from collections import OrderedDict

NEO_SC_FRAMEWORK = 'boa.blockchain.vm.'

import pdb

class VMToken():
    """

    """
    addr = None

    pytoken = None

    data = None

    vm_op = None

    src_method = None

    target_method = None

    is_annotation = None

    updatable_data = None

    @property
    def out_op(self):
        """

        :return:
        """
        if type(self.vm_op) is int:
            return self.vm_op
        elif type(self.vm_op) is bytes:
            return ord(self.vm_op)
        else:
            raise Exception('Invalid op: %s ' % self.vm_op)

    def __init__(self, vm_op=None, pytoken=None, addr=None, data=None):
        self.vm_op = vm_op
        self.pytoken = pytoken
        self.addr = addr

        if self.pytoken is not None and hasattr(self.pytoken, 'data'):
            self.data = self.pytoken.data

        self.data = data

        self.src_method = None
        self.target_method = None

        self.is_annotation = False


class VMTokenizer():
    """

    """
    method = None

    _address = None

    vm_tokens = None

    total_param_and_body_count_token = None

    def __init__(self, method):
        self.method = method
        self._address = 0
        self.vm_tokens = OrderedDict()

        self.method_begin_items()

    def to_s(self):

        """

        """
        lineno = self.method.start_line_no
        pstart = True
        for i, (key, value) in enumerate(self.vm_tokens.items()):

            if value.pytoken:
                pt = value.pytoken

                do_print_line_no = False
                to_label = None
                from_label = '    '
                if pt.line_no != lineno:
                    print("\n")
                    lineno = pt.line_no
                    do_print_line_no = True

                if pt.args and type(pt.args) is Label:
                    addr = value.addr
                    if value.data is not None:
                        plus_addr = int.from_bytes(
                            value.data, 'little', signed=True)
                        target_addr = addr + plus_addr
                        to_label = 'to %s    [ %s ]' % (target_addr, pt.args)
                    else:
                        to_label = 'from << %s ' % pt.args
                        #                    to_label = 'to %s ' % pt.args
                elif pt.jump_label:
                    from_label = ' >> '
                    to_label = 'from [%s]' % pt.jump_label

                ds = ''
                if value.data is not None:
                    try:
                        ds = int.from_bytes(value.data, 'little', signed=True)
                    except Exception as e:
                        pass
                    if type(ds) is not int and len(ds) < 1:
                        try:
                            ds = value.data.decode('utf-8')
                        except Exception as e:
                            pass

                if pt.py_op == pyop.CALL_FUNCTION:
                    to_label = '%s %s ' % (pt.func_name, pt.func_params)

                lno = "{:<10}".format(
                    pt.line_no if do_print_line_no or pstart else '')
                addr = "{:<4}".format(key)
                op = "{:<20}".format(str(pt.py_op))
                arg = "{:<50}".format(
                    to_label if to_label is not None else pt.arg_s)
                data = "[data] {:<20}".format(ds)
                print("%s%s%s%s%s%s" % (lno, from_label, addr, op, arg, data))

            pstart = False

    def to_b(self):
        """

        :return:
        """
        b_array = bytearray()
        for key, vm_token in self.vm_tokens.items():

            b_array.append(vm_token.out_op)

            if vm_token.data is not None and vm_token.vm_op != VMOp.NOP:
                b_array = b_array + vm_token.data

        return b_array

    def method_begin_items(self):

        # we just need to inssert the total number of arguments + body variables
        # which is the length of the method `local_stores` dictionary
        # then create a new array for the vm to store
        """

        """
        total_items = self.method.total_lines \
                      + len(self.method.args) \
                      + self.method.dynamic_iterator_count

        self.total_param_and_body_count_token = self.insert_push_integer(
            total_items)
        self.total_param_and_body_count_token.updatable_data = total_items
        self.insert1(VMOp.NEWARRAY)
        self.insert1(VMOp.TOALTSTACK)

        for index, arg in enumerate(self.method.args):
            self.convert_load_parameter(arg, index)

    def update_method_begin_items(self):
        """

        """
        num_current_items = self.total_param_and_body_count_token.updatable_data

        if self.method.dynamic_iterator_count > 0:
            num_current_items += self.method.dynamic_iterator_count
            self.update_push_integer(
                self.total_param_and_body_count_token, num_current_items)

    def insert_vm_token_at(self, vm_token, index):

        """

        :param vm_token:
        :param index:
        """
        self.vm_tokens[index] = vm_token

    def update1(self, vmtoken, vm_op, data=None):

        """

        :param vmtoken:
        :param vm_op:
        :param data:
        :return:
        """
        vmtoken.vm_op = vm_op
        vmtoken.data = data

        self.insert_vm_token_at(vmtoken, vmtoken.addr)

        return vmtoken

    def update_push_data(self, vmtoken, data):

        """

        :param vmtoken:
        :param data:
        :return:
        """
        dlen = len(data)

        if dlen == 0:
            return self.update1(vmtoken, VMOp.PUSH0)

        elif dlen <= 75:
            return self.update1(vmtoken, dlen, data)

        if dlen < 0x100:
            prefixlen = 1
            code = VMOp.PUSHDATA1

        elif dlen < 0x1000:
            prefixlen = 2
            code = VMOp.PUSHDATA2

        else:
            prefixlen = 4
            code = VMOp.PUSHDATA4

        byts = bytearray(dlen.to_bytes(prefixlen, 'little')) + data

        return self.update1(vmtoken, code, byts)

    def update_push_integer(self, vmtoken, i):
        """

        :param vmtoken:
        :param i:
        :return:
        """
        if i == 0:
            return self.update1(vmtoken, VMOp.PUSH0)
        elif i == -1:
            return self.insert1(vmtoken, VMOp.PUSHM1)
        elif i > 0 and i <= 16:
            out = 0x50 + i
            return self.update1(vmtoken, out)

        bigint = BigInteger(i)
        outdata = bigint.ToByteArray()

        return self.update_push_data(vmtoken, outdata)

    def insert1(self, vm_op, data=None):

        """

        :param vm_op:
        :param data:
        :return:
        """
        start_addr = self._address

        vmtoken = VMToken(vm_op=vm_op, addr=start_addr, data=data)

        self._address += 1

        if vmtoken.data is not None:
            self._address += len(vmtoken.data)

        self.insert_vm_token_at(vmtoken, vmtoken.addr)

        return vmtoken

    def insert_push_data(self, data):

        """

        :param data:
        :return:
        """
        dlen = len(data)

        if dlen == 0:
            return self.insert1(VMOp.PUSH0)

        elif dlen <= 75:
            return self.insert1(dlen, data)

        if dlen < 0x100:
            prefixlen = 1
            code = VMOp.PUSHDATA1

        elif dlen < 0x1000:
            prefixlen = 2
            code = VMOp.PUSHDATA2

        else:
            prefixlen = 4
            code = VMOp.PUSHDATA4

        byts = bytearray(dlen.to_bytes(prefixlen, 'little')) + data

        return self.insert1(code, byts)

    def insert_push_integer(self, i):
        """

        :param i:
        :return:
        """
        if i == 0:
            return self.insert1(VMOp.PUSH0)
        elif i == -1:
            return self.insert1(VMOp.PUSHM1)
        elif i > 0 and i <= 16:
            out = 0x50 + i
            return self.insert1(out)

        bigint = BigInteger(i)
        outdata = bigint.ToByteArray()

        return self.insert_push_data(outdata)

    def convert1(self, vm_op, py_token=None, data=None):

        """

        :param vm_op:
        :param py_token:
        :param data:
        :return:
        """
        start_addr = self._address

        vmtoken = VMToken(vm_op=vm_op, addr=start_addr,
                          pytoken=py_token, data=data)

        self._address += 1

        if vmtoken.data is not None and type(vmtoken.data) is not Label:
            self._address += len(data)

        self.insert_vm_token_at(vmtoken, start_addr)

        return vmtoken

    def convert_new_array(self, vm_op, py_token=None, data=None):

        # push the length of the array
        """

        :param vm_op:
        :param py_token:
        :param data:
        """
        if type(py_token.args) is int:

            self.insert_push_integer(py_token.args)
        else:
            self.convert_load_local(py_token, py_token.args)

        self.convert1(VMOp.PACK, py_token)

    def convert_push_data(self, data, py_token=None):

        """

        :param data:
        :param py_token:
        :return:
        """
        dlen = len(data)
        if dlen == 0:
            return self.convert1(VMOp.PUSH0, py_token=py_token)
        elif dlen <= 75:
            return self.convert1(len(data), py_token=py_token, data=data)

        if dlen < 0x100:
            prefixlen = 1
            code = VMOp.PUSHDATA1
        elif dlen < 0x1000:
            prefixlen = 2
            code = VMOp.PUSHDATA2
        else:
            prefixlen = 4
            code = VMOp.PUSHDATA4

        byts = bytearray(dlen.to_bytes(prefixlen, 'little')) + data

        return self.convert1(code, py_token=py_token, data=byts)

    def convert_push_integer(self, i, py_token=None):
        """

        :param i:
        :param py_token:
        :return:
        """
        if i == 0:
            return self.convert1(VMOp.PUSH0, py_token=py_token)
        elif i == -1:
            return self.convert1(VMOp.PUSHM1, py_token=py_token)
        elif i > 0 and i <= 16:
            out = 0x50 + i
            return self.convert1(out, py_token=py_token)

        bigint = BigInteger(i)

        outdata = bigint.ToByteArray()

        return self.convert_push_data(outdata, py_token=py_token)

    def convert_store_local(self, py_token):

        # set array
        """

        :param py_token:
        """
        self.convert1(VMOp.FROMALTSTACK, py_token=py_token)
        self.convert1(VMOp.DUP)
        self.convert1(VMOp.TOALTSTACK)

        local_name = py_token.args

        position = self.method.local_stores[local_name]

        # set i the index of the local variable to be stored
        self.convert_push_integer(position)

        # set item
        self.convert_push_integer(2)
        self.convert1(VMOp.ROLL)
        self.convert1(VMOp.SETITEM)

    def convert_load_local(self, py_token, name=None):

        """

        :param py_token:
        :param name:
        """
        if name is not None:
            local_name = name
        else:
            local_name = py_token.args


        #check to see if this local is a variable
        if local_name in self.method.local_stores:

            position = self.method.local_stores[local_name]

            # get array
            self.convert1(VMOp.FROMALTSTACK, py_token=py_token)
            self.convert1(VMOp.DUP)
            self.convert1(VMOp.TOALTSTACK)

            # get i
            self.convert_push_integer(position)
            self.convert1(VMOp.PICKITEM)

        else:
            py_token.func_params = []
            py_token.func_name = local_name

            self.convert_method_call(py_token)


    def insert_unknown_type(self, item):
        """

        :param item:
        """
        if type(item) is int:
            self.insert_push_integer(item)

        elif type(item) is str:
            str_bytes = item.encode('utf-8')
            self.insert_push_data(str_bytes)

        elif type(item) is bytearray:
            self.insert_push_data(bytes(item))

        elif type(item) is bytes:
            self.insert_push_data(item)

        elif type(item) is bool:
            self.insert_push_data(item)
        elif type(item) == type(None):
            self.insert_push_data(bytearray(0))
        else:
            raise Exception("Could not load type %s for item %s " %
                            (type(item), item))

    def convert_set_element(self, arg, position):

        #        print("converting set element %s %s" % (position, type(position)))

        """

        :param arg:
        :param position:
        """
        if type(position) is int:

            self.insert_push_integer(position)
        elif type(position) is str:
            self.convert_load_local(None, name=position)

        if type(arg.array_item) is str:

            # first we'll look for the local variable with name of the str
            if arg.array_item in self.method.local_stores:
                self.convert_load_local(None, name=arg.array_item)
            # otherwise we'll do the unknown type thing
            else:
                self.insert_unknown_type(arg.array_item)
        else:
            self.insert_unknown_type(arg.array_item)

        self.convert1(VMOp.SETITEM, arg)

    def convert_load_parameter(self, arg, position):

        """

        :param arg:
        :param position:
        """
        length = len(self.method.local_stores)
        self.method.local_stores[arg] = length

        # get array
        self.insert1(VMOp.FROMALTSTACK)
        self.insert1(VMOp.DUP)
        self.insert1(VMOp.TOALTSTACK)

        self.insert_push_integer(position)
        self.insert_push_integer(2)

        self.insert1(VMOp.ROLL)
        self.insert1(VMOp.SETITEM)

    def convert_built_in_list(self, pytoken):
        """

        :param pytoken:
        """
        new_array_len = 0
        lenfound = False
        for index, token in enumerate(pytoken.func_params):

            if token.args == 'length' and not lenfound:
                # first we see if a constant ( ie integer was passed in

                new_array_len = pytoken.func_params[index + 1].args

                if type(new_array_len) is int:
                    self.insert_push_integer(new_array_len)
                else:
                    self.convert_load_local(None, name=new_array_len)
                lenfound = True

        if not lenfound:
            self.insert_push_integer(0)
        self.convert1(VMOp.NEWARRAY, pytoken)

    def convert_build_slice(self, pytoken):

        # this was fun!

        # rotate so list is on the top, then move it to alt stack
        self.convert1(VMOp.ROT)
        self.convert1(VMOp.TOALTSTACK, py_token=pytoken)

        # swap the end index and the start index, duplicate start index to alt stack
        self.convert1(VMOp.SWAP)
        self.convert1(VMOp.DUP)
        self.convert1(VMOp.TOALTSTACK)

        # subtract end index from start index, this is placed on the stack
        self.convert1(VMOp.SUB)

        # get the start index and list from alt stack
        self.convert1(VMOp.FROMALTSTACK)
        self.convert1(VMOp.FROMALTSTACK)

        # swap the list and the start index
        self.convert_push_integer(2)
        self.convert1(VMOp.XSWAP)

        # and now perform substr. whew.
        self.convert1(VMOp.SUBSTR)


    def convert_method_call(self, pytoken):

        # special case for list initialization
        """

        :param pytoken:
        :return:
        """
        if pytoken.func_name == 'list':
            return self.convert_built_in_list(pytoken)
        elif pytoken.func_name == 'bytearray':
            return self.convert_push_data(bytes(pytoken.func_params[0].args), pytoken)
        elif pytoken.func_name == 'bytes':
            return self.convert_push_data(pytoken.func_params[0].args, pytoken)
        elif pytoken.func_name == 'AppCall':
            scripthash_token = pytoken.func_params.pop(0)
            pytoken.script_hash_token = scripthash_token.args

        for t in pytoken.func_params:
            t.to_vm(self)

        param_len = len(pytoken.func_params)

        if param_len <= 1:
            pass
        elif param_len == 2:
            # if we are using concat or take, we don't want to swap
            if pytoken.func_name != 'concat' and pytoken.func_name != 'take':
                self.insert1(VMOp.SWAP)

        elif param_len == 3:

            if pytoken.func_name != 'substr':
                self.insert_push_integer(2)
                self.insert1(VMOp.XSWAP)

        else:
            half_p = int(param_len / 2)

            for i in range(0, half_p):
                save_to = param_len - 1 - i

                self.insert_push_integer(save_to)
                self.insert1(VMOp.PICK)

                self.insert_push_integer(i + 1)
                self.insert1(VMOp.PICK)

                self.insert_push_integer(save_to + 2)
                self.insert1(VMOp.XSWAP)
                self.insert1(VMOp.DROP)

                self.insert_push_integer(i + 1)
                self.insert1(VMOp.XSWAP)
                self.insert1(VMOp.DROP)

        self.insert1(VMOp.NOP)

        fname = pytoken.func_name
        full_name = None
        for m in self.method.module.methods:
            if fname == m.name:
                full_name = m.full_name

        # operational call like len(items) or abs(value)
        if self.is_op_call(fname):
            vmtoken = self.convert_op_call(fname, pytoken)

        # runtime.notify event
        elif self.is_notify_event(pytoken):
            vmtoken = self.convert_notify_event(pytoken)

        # app call ( for calling other contracts on blockchain )
        elif self.is_smart_contract_call(pytoken):
            vmtoken = self.convert_smart_contract_call(pytoken)

        elif self.is_sys_call(full_name):
            vmtoken = self.convert_sys_call(full_name, pytoken)

        # used for python specific built in methods like `enumerate` or `tuple`
        elif self.is_built_in(fname):
            vmtoken = self.convert_built_in(fname, pytoken)

        # otherwise we assume the method is defined by the module
        else:
            vmtoken = self.convert1(
                VMOp.CALL, py_token=pytoken, data=bytearray(b'\x05\x00'))

            vmtoken.src_method = self.method
            vmtoken.target_method = pytoken.func_name

        return vmtoken

    def is_op_call(self, op):
        """

        :param op:
        :return:
        """
        if op in ['len', 'abs', 'min', 'max', 'concat', 'take', 'substr',
                  'sha1', 'sha256', 'hash160', 'hash256',
                  'verify_signature', 'verify_signatures']:
            return True
        return False

    def convert_op_call(self, op, pytoken=None):

        """

        :param op:
        :param pytoken:
        :return:
        """
        if op == 'len':
            return self.convert1(VMOp.ARRAYSIZE, pytoken)
        elif op == 'abs':
            return self.convert1(VMOp.ABS, pytoken)
        elif op == 'min':
            return self.convert1(VMOp.MIN, pytoken)
        elif op == 'max':
            return self.convert1(VMOp.MAX, pytoken)
        elif op == 'concat':
            return self.convert1(VMOp.CAT, pytoken)
        elif op == 'take':
            return self.convert1(VMOp.LEFT, pytoken)
        elif op == 'substr':
            return self.convert1(VMOp.SUBSTR, pytoken)
        elif op == 'sha1':
            return self.convert1(VMOp.SHA1, pytoken)
        elif op == 'sha256':
            return self.convert1(VMOp.SHA256, pytoken)
        elif op == 'hash160':
            return self.convert1(VMOp.HASH160, pytoken)
        elif op == 'hash256':
            return self.convert1(VMOp.HASH256, pytoken)
        elif op == 'verify_signature':
            return self.convert1(VMOp.CHECKSIG, pytoken)
        elif op == 'verify_signatures':
            return self.convert1(VMOp.CHECKMULTISIG, pytoken)
        return None

    def is_sys_call(self, op):
        """

        :param op:
        :return:
        """
        if op is not None and NEO_SC_FRAMEWORK in op:
            if 'TriggerType' not in op:  # we will compile TriggerType normally
                return True
        return False

    def convert_sys_call(self, op, pytoken=None):

        """

        :param op:
        :param pytoken:
        :return:
        """
        syscall_name = op.replace(NEO_SC_FRAMEWORK, '').encode('utf-8')
        length = len(syscall_name)
        ba = bytearray([length]) + bytearray(syscall_name)
        pytoken.is_sys_call = False
        vmtoken = self.convert1(VMOp.SYSCALL, pytoken, data=ba)
        self.insert1(VMOp.NOP)
        return vmtoken

    def is_built_in(self, op):

        """

        :param op:
        :return:
        """
        if op in ['zip', 'type', 'tuple', 'super', 'str', 'slice',
                  'set', 'reversed', 'property', 'memoryview',
                  'map', 'list', 'frozenset', 'float', 'filter',
                  'enumerate', 'dict', 'divmod', 'complex', 'bytes', 'bytearray', 'bool',
                  'int', 'vars', 'sum', 'sorted', 'round', 'setattr', 'getattr',
                  'rep', 'quit', 'print', 'pow', 'ord', 'oct', 'next', 'locals', 'license',
                  'iter', 'isinstance', 'issubclass', 'input', 'id', 'hex',
                  'help', 'hash', 'hasattr', 'globals', 'format', 'exit',
                  'exec', 'eval', 'dir', 'deleteattr', 'credits', 'copyright',
                  'compile', 'chr', 'callable', 'bin', 'ascii', 'any', 'all', ]:
            return True

        return False

    def convert_built_in(self, op, pytoken):

        """

        :param op:
        :param pytoken:
        :return:
        """
        if op == 'print':
            syscall_name = 'Neo.Runtime.Log'.encode('utf-8')
            length = len(syscall_name)
            ba = bytearray([length]) + bytearray(syscall_name)
            vmtoken = self.convert1(VMOp.SYSCALL, pytoken, data=ba)
            self.insert1(VMOp.NOP)
            return vmtoken

        raise NotImplementedError(
            "[Compilation error] Built in %s is not implemented" % op)

    def is_notify_event(self, pytoken):

        """

        :param pytoken:
        :return:
        """
        name = pytoken.func_name

        for action in self.method.module.actions:
            if action.method_name == name:
                return True
        return False

    def convert_notify_event(self, pytoken):
        """

        :param pytoken:
        :return:
        """
        event_action = None
        for action in self.method.module.actions:
            if action.method_name == pytoken.func_name:
                event_action = action
        if event_action is None:
            raise Exception("Event action not found")

        # push the event name
        event_name = event_action.event_name.encode('utf-8')
        self.convert_push_data(event_name, py_token=pytoken)

        # push the num params
        self.convert_push_integer(len(event_action.event_args))

        # pack the array
        self.convert1(VMOp.PACK)

        # insert syscall
        syscall_name = 'Neo.Runtime.Notify'.encode('utf-8')
        length = len(syscall_name)
        ba = bytearray([length]) + bytearray(syscall_name)
        vmtoken = self.convert1(VMOp.SYSCALL, pytoken, data=ba)

        return vmtoken

    def is_smart_contract_call(self, pytoken):
        """

        :param pytoken:
        :return:
        """
        name = pytoken.func_name


        if name == 'AppCall':
            return True


        for appcall in self.method.module.app_call_registrations:
            if appcall.method_name == name:
                return True
        return False

    def convert_smart_contract_call(self, pytoken):
        """

        :param pytoken:
        :return:
        """

        # the following converts app calls of the pattern
        # m = AppCall(script_hash, *args)
        if pytoken.script_hash_addr is not None:

            from boa.code.items import SmartContractAppCall

            shash = SmartContractAppCall.ToScriptHashData( pytoken.script_hash_addr )

            vmtoken = self.convert1(VMOp.APPCALL, py_token=pytoken, data=shash)
            return vmtoken

        # this is used for app calls that are registered
        # using RegisterAppCall(script_hash, *args)
        sc_appcall = None
        for appcall in self.method.module.app_call_registrations:
            if appcall.method_name == pytoken.func_name:
                sc_appcall = appcall
        if sc_appcall is None:
            raise Exception("Smart Contract Appcall %s not found " %
                            pytoken.func_name)

        # push the contract hash
        vmtoken = self.convert1(
            VMOp.APPCALL, py_token=pytoken, data=sc_appcall.script_hash_addr)

        return vmtoken



from byteplay3 import *

from boa.code import pyop


class Line():

    """

    """
    items = None

    def __init__(self, item_list):
        self.items = item_list

    @property
    def is_import(self):
        """

        :return:
        """
        for i, (op, arg) in enumerate(self.items):
            if op in [pyop.IMPORT_NAME, pyop.IMPORT_FROM, pyop.IMPORT_STAR]:
                return True
        return False

    @property
    def is_definition(self):
        """

        :return:
        """
        return len(self.items) == 3 and self.items[1][0] == pyop.LOAD_CONST and self.items[2][0] == pyop.STORE_NAME
#        return False

    @property
    def is_docstring(self):
        """
        returns whether a line is a docstring

        :return: whether a line is a documentation string
        :rtype: bool

        """
        for item in self.items:
            if item[0] == pyop.STORE_NAME and item[1] == '__doc__':
                return True
        return False



    @property
    def is_method(self):
        """

        :return:
        """
        for i, (op, arg) in enumerate(self.items):
            if op == pyop.MAKE_FUNCTION:
                return True
        return False

    @property
    def is_class(self):
        """

        :return:
        """
        for i, (op, arg) in enumerate(self.items):
            if op == pyop.LOAD_BUILD_CLASS:
                return True
        return False

    @property
    def code_object(self):
        """

        :return:
        """
        for i, (op, arg) in enumerate(self.items):
            if type(arg) is Code:
                return arg
        return None

    @property
    def is_action_registration(self):
        """

        :return:
        """
        for i, (op, arg) in enumerate(self.items):
            if arg == 'RegisterAction':
                return True

    @property
    def is_smart_contract_appcall_registration(self):
        """

        :return:
        """
        for i, (op, arg) in enumerate(self.items):
            if arg == 'RegisterAppCall':
                return True




class Account():

    @property
    def ScriptHash(self):
        """

        :return:
        """
        return GetScriptHash(self)

    @property
    def Votes(self):
        """

        :return:
        """
        return GetVotes(self)


def GetScriptHash(account):
    """

    :param account:
    """
    pass


def GetVotes(account):
    """

    :param account:
    """
    pass


def SetVotes(account, votes):
    """

    :param account:
    :param votes:
    """
    pass


def GetBalance(account, asset_id):
    """

    :param account:
    :param asset_id:
    """
    pass



from byteplay3 import Code, SetLinenoType, Label
from boa.code import pyop

from boa.code.line import Line
from boa.code.method import Method
from boa.code.items import Definition, Klass, Import, Action, SmartContractAppCall

from boa.blockchain.vm import VMOp

from collections import OrderedDict

import pdb


class Module():
    """
    A Module is the top level component which contains code objects.
    When, for example, compiling ``path/to/my/file.py``, the items contained in ``file.py`` are the module.
    An executable may have many modules.  The 'default' or 'entry' module in the example above would be ``file.py``

    When calling ``Compiler.load_and_save('path/to/file.py')``, a module is created for ``file.py``.
    If `file.py` imports any other functionality, those modules will also be added to the executable
    and placed in the Module.loaded_modules attribute.

    After modules have been processed as methods, and then methods processed as blocks, and blocks processed to tokens,
    The main or ``default`` module's ``write()`` method is called, which writes the executable to a byte string and returns
    it to be saved to disk.

    If you would like to inspect the contents of a module, you may use the ``Compiler.load('path/to/file.py')`` which will
    return an instance of the compiler.  With that instance, you can access the ``default`` module of the compiler
    which will in turn give you access to any other loaded modules contained in the default module.


    Each module ( as well as each method object ) contains a reference to a ``byteplay3`` object, named ``bp``.
    This object contains the instruction set as it would be viewed in the Python interpreter.

    You can call ``print(module.bp.code)`` on any object with a ``bp`` attribute, and it will output the Python interpreter code.


    >>> from boa.compiler import Compiler
    >>> module = Compiler.load('./boa/tests/src/AddTest1.py').default
    >>> print(module.bp.code)
    2     1 LOAD_CONST           <byteplay3.Code object at 0x10cc3d6a0>
          2 LOAD_CONST           'Main'
          3 MAKE_FUNCTION        0
          4 STORE_NAME           Main
          5 LOAD_CONST           None
          6 RETURN_VALUE


    Once an executable has been processed and tokenized, it will then have a set of vm tokens that are similar
    to the ``byteplay3`` tokens, but different in important ways. These are contained in the module's ``all_vm_tokens`` attribute

    You may call ``module.to_s()`` to view the program as it has been tokenized for the NEO Virtual Machine.


    >>> module.to_s()
    4             31  LOAD_FAST           a                [data]
                  36  LOAD_CONST          2                [data]
                  37  BINARY_MULTIPLY                      [data]
                  38  STORE_FAST          a2               [data]
    6             45  LOAD_FAST           b                [data]
                  50  LOAD_CONST          1                [data]
                  51  BINARY_ADD                           [data]
                  52  STORE_FAST          b2               [data]
    8             59  LOAD_FAST           c                [data]
                  64  LOAD_CONST          2                [data]
                  65  BINARY_TRUE_DIVIDE                   [data]
                  66  STORE_FAST          c2               [data]
    10            73  LOAD_FAST           d                [data]
                  78  LOAD_CONST          1                [data]
                  79  BINARY_SUBTRACT                      [data]
                  80  STORE_FAST          d2               [data]
    13            87  243                 b'\x03\x00'      [data] 3
                  90  LOAD_FAST           a2               [data]
                  95  LOAD_FAST           b2               [data]
                  100 BINARY_ADD                           [data]
                  101 LOAD_FAST           c2               [data]
                  106 BINARY_ADD                           [data]
                  107 LOAD_FAST           d2               [data]
                  112 BINARY_ADD                           [data]
                  113 NOP                                  [data]
                  114 241                                  [data]
                  115 242                                  [data]
                  116 RETURN_VALUE                         [data]
    """
    
    bp = None  # this is to store the byteplay reference

    path = None  # the path where this file is

    lines = None  # this contains the code objects split up into different line start indexes

    imports = None  # a list of import statements

    module_variables = None  # list of module variables

    classes = None  # a list of classes

    methods = None  # a list to keep all methods in the module

    actions = None  # a list to keep track of event registrations

    app_call_registrations = None  # a list to keep track of app call registrations

    is_sys_module = None

    all_vm_tokens = None # dict for converting method tokens into linked method tokens for writing

    loaded_modules = None

    _module_name = None

    _names_to_load = None


    @property
    def module_path(self):
        """
        Return the file path of the module.

        :return: the path of the module
        :rtype: str
        """
        
        return self._module_name
        

    @property
    def main(self):
        """
        Return the default method in this module.

        :return: the default method in this module
        :rtype: ``boa.code.method.Method``
        """
        
        for m in self.methods:
            
            if m.name == 'Main':
                return m
                
        if len(self.methods):
            return self.methods[0]
            
        return None
        

    @property
    def orderered_methods(self):
        """
        An ordered list of methods

        :return: A list of ordered methods is this module
        :rtype: list
        """
        
        oms = []
        
        self.methods.reverse()
        
        if self.main:
            oms = [self.main]

        for m in self.methods:
            
            if m == self.main:
                continue
            oms.append(m)

        return oms


    def add_method(self, method):
        """
        Add a method to this module.

        :param method: the method object to add to this module
        :type method: ``boa.code.method.Method``

        :return: whether the method was added
        :rtype: bool
        """
        
        for m in self.methods:
            
            if m.name == method.name:
    
                if m.name != m.full_name:
                    
                    if m.full_name == method.full_name:
                        return False
                else:
                    return False
                # return False

        # print("appending method %s %s " % (method.name, method.full_name))
        self.methods.append(method)
        
        return True


    def method_by_name(self, method_name):
        """
        Look up a method by its name from the module ``methods`` list.
        :param method_name: the name of the method to look up
        :type method_name: str

        :return: the method ( if it is found)
        :rtype: ``boa.code.method.Method``
        """
        
        for m in self.methods:
            if m.full_name == method_name:
                return m
            elif m.name == method_name:
                return m
        return None
        

    def __init__(self, path, module_name='', is_sys_module=False, items_to_import=None):

        self.path = path

        self._module_name = module_name

        self.is_sys_module = is_sys_module
        
        self._names_to_load = ['STAR'] if items_to_import is None else items_to_import

        source = open(path, 'rb')

        suite = compile(source.read(), path, 'exec')

        self.bp = Code.from_code(suite)

        source.close()

        self.build()
        

    def build(self):
        """
        Split the ``bp.code`` object into lines, and assembles the lines into different items.
        """
        
        self.lines = []
        self.imports = []
        self.module_variables = []
        self.methods = []
        self.actions = []
        self.app_call_registrations = []
        self.classes = []
        self.loaded_modules = []

        self.split_lines()

        for lineset in self.lines:

            if lineset.is_import:

                if not self.is_sys_module:
                    imp = Import(lineset.items)
                    self.process_import(imp)
                else:
                    print("will not import items from sys module")

            elif lineset.is_docstring:
                pass
            elif lineset.is_definition:
                self.module_variables.append(Definition(lineset.items))
            elif lineset.is_class:
                self.classes.append(Klass(lineset.items, self))
            elif lineset.is_method:
                self.process_method(lineset)
            elif lineset.is_action_registration:
                self.process_action(lineset)
            elif lineset.is_smart_contract_appcall_registration:
                self.process_smart_contract_app_registration(lineset)
            else:
                print('not sure what to do with line %s ' % lineset)
                pdb.set_trace()


    def process_import(self, import_item):
        """
        Processes an import statement within this module.

        :param import_item:
        :type import_item: ``boa.code.items.Import subclass``
        """
        
        self.imports.append(import_item)

        self.loaded_modules.append(import_item.imported_module)

        # Go through all the methods in the imported module.
        for method in import_item.imported_module.methods:
            self.add_method(method)


    def process_method(self, lineset):
        """
        processes a set of lines that contain a byteplay3 code object

        :param lineset: the lineset to process and add
        :type lineset: list
        """
        
        m = Method(lineset.code_object, self)

        if 'STAR' in self._names_to_load:
            self.add_method(m)
        else:
            for item in self._names_to_load:
                
                if item == m.name:
                    self.add_method(m)

    def process_action(self, lineset):
        """
        Processes an action within this module.
        A sample action would be to create an event like so:

        .. code-block:: python

            from boa.blockchain.vm.Neo.Action import RegisterAction

            # Register the action.
            onRefund = RegisterAction('refund', 'to_address', 'amount')

            # Dispatch an action.
            onRefund(my_address, 100)

        :param lineset: The set of lines containing an app call registration
        :type lineset: list
        """
        
        action = Action(lineset)
        for act in self.actions:
            if act.method_name == action.method_name:
                return None
                
        self.actions.append(action)


    def process_smart_contract_app_registration(self, lineset):
        """
        processes a smart contract app registration, for when you would like to call another
        smart contract from your contract.  for example:

        .. code-block:: python

            from boa.blockchain.vm.Neo.App import RegisterAppCall

            # register the contract
            otherContract = RegisterAppCall('contract_hash', 'param1','param2')

            # call the contract
            result = otherContract( a, b )

        :param lineset: The set of lines containing an app call registration
        :type lineset: list
        """
        
        appcall_registration = SmartContractAppCall(lineset)
        for registration in self.app_call_registrations:
            if registration.method_name == appcall_registration.method_name:
                return
        self.app_call_registrations.append(appcall_registration)


    def split_lines(self):
        """
        Split the list of lines in the module into a set of objects that can be interpreted.
        """
        
        lineitem = None

        for i, (op, arg) in enumerate(self.bp.code):

            if isinstance(op, SetLinenoType):
                if lineitem is not None:
                    self.lines.append(Line(lineitem))

                lineitem = []

            lineitem.append((op, arg))

        if len(lineitem):
            self.lines.append(Line(lineitem))


    def write(self):
        """
        Write the current module to a byte string.

        Note that if you are using the ``Compiler.load('path/to/file.py')``, you must 
        call ``module.write()`` before any inspection of the module is possible.

        :return: A bytestring of representing the current module
        :rtype: bytes
        """
        self.link_methods()

        return self.write_methods()


    def write_methods(self):
        """
        Write all methods in the current module to a byte string.

        :return: A bytestring of all current methods in this module
        :rtype: bytes
        """

        b_array = bytearray()
        for key, vm_token in self.all_vm_tokens.items():

            b_array.append(vm_token.out_op)

            if vm_token.data is not None and vm_token.vm_op != VMOp.NOP:
                b_array = b_array + vm_token.data

        return b_array
        

    def link_methods(self):
        """
        Perform linkage of addresses between methods.
        """

        self.all_vm_tokens = OrderedDict()

        address = 0

        for method in self.orderered_methods:
            method.method_address = address

            for key, vmtoken in method.vm_tokens.items():
                self.all_vm_tokens[address] = vmtoken

                address += 1

                if vmtoken.data is not None:
                    address += len(vmtoken.data)

                vmtoken.addr = vmtoken.addr + method.method_address

        for key, vmtoken in self.all_vm_tokens.items():
            if vmtoken.src_method is not None:

                target_method = self.method_by_name(vmtoken.target_method)
                if target_method:

                    jump_len = target_method.method_address - vmtoken.addr
                    vmtoken.data = jump_len.to_bytes(2, 'little', signed=True)
                else:
                    raise Exception("Target method %s not found" % vmtoken.target_method)


    def to_s(self):
        """
        this method is used to print the output of the executable in a readable/ tokenized format.
        sample usage:

        >>> from boa.compiler import Compiler
        >>> module = Compiler.load('./boa/tests/src/LambdaTest.py').default
        >>> module.write()
        >>> module.to_s()
        12            3   LOAD_CONST          9                [data]
                      4   STORE_FAST          j                [data]
        22            11  LOAD_FAST           j                [data]
                      17  CALL_FUNCTION       Main.<locals>.q_1 \
					  [<boa.code.pytoken.PyToken object at 0x10cb53c50>] [data] 22
                      20  STORE_FAST          m                [data]
        24            27  243                 b'\x03\x00'      [data] 3
                      30  LOAD_FAST           m                [data]
                      35  NOP                                  [data]
                      36  241                                  [data]
                      37  242                                  [data]
                      38  RETURN_VALUE                         [data]
        20            49  243                 b'\x03\x00'      [data] 3
                      52  LOAD_FAST           x                [data]
                      57  LOAD_CONST          1                [data]
                      58  BINARY_ADD                           [data]
                      59  NOP                                  [data]
                      60  241                                  [data]
                      61  242                                  [data]
                      62  RETURN_VALUE                         [data]
        """

        lineno = 0
        pstart = True
        
        for i, (key, value) in enumerate(self.all_vm_tokens.items()):
            if value.pytoken:
                pt = value.pytoken
                do_print_line_no = False
                to_label = None
                from_label = '    '
                
                if pt.line_no != lineno:
                    print("\n")
                    lineno = pt.line_no
                    do_print_line_no = True

                if pt.args and type(pt.args) is Label:
                    addr = value.addr
                    if value.data is not None:
                        plus_addr = int.from_bytes(
                            value.data, 'little', signed=True)
                        target_addr = addr + plus_addr
                        to_label = 'to %s    [ %s ]' % (target_addr, pt.args)
                    else:
                        to_label = 'from << %s ' % pt.args
                        #                    to_label = 'to %s ' % pt.args
                elif pt.jump_label:
                    from_label = ' >> '
                    to_label = 'from [%s]' % pt.jump_label

                ds = ''
                if value.data is not None:
                    try:
                        ds = int.from_bytes(value.data, 'little', signed=True)
                    except Exception as e:
                        pass
                    if type(ds) is not int and len(ds) < 1:
                        try:
                            ds = value.data.decode('utf-8')
                        except Exception as e:
                            pass

                if pt.py_op == pyop.CALL_FUNCTION:
                    to_label = '%s %s ' % (pt.func_name, pt.func_params)

                lno = "{:<10}".format(
                    pt.line_no if do_print_line_no or pstart else '')
                addr = "{:<4}".format(key)
                op = "{:<20}".format(str(pt.py_op))
                arg = "{:<50}".format(
                    to_label if to_label is not None else pt.arg_s)
                data = "[data] {:<20}".format(ds)
                print("%s%s%s%s%s%s" % (lno, from_label, addr, op, arg, data))

            pstart = False



from boa.code.builtins import range


def Main():

    """

    :return:
    """
    count = 0

    for i in range(0, 5):
        count = count + i
        count += awesome()
        count += not_so_awesome()

    return count


def awesome():

    """

    :return:
    """
    return 2


def not_so_awesome():
    """

    :return:
    """
    return -1



from boa.code.builtins import concat
from boa.blockchain.vm.Neo.Runtime import Notify


def Main():

    """

    :return:
    """
    mystring = 'abcdefgh'
    count = 3
    mys = mystring[0:count]

    strmy = mystring[4:]

    Notify(mys)
    Notify(strmy)

    myarray = [b'\x10', b'\x20', b'\x30', b'\x40']

    # mr1 = myarray[0:2] this does not work, need to implement a bytearray thing of some sort
    # Notify(mr1)

    return 1




class TransactionAttribute():

    @property
    def Usage(self):
        """

        :return:
        """
        return GetUsage(self)

    @property
    def Data(self):
        """

        :return:
        """
        return GetData(self)


def GetUsage(transaction_attr):
    """

    :param transaction_attr:
    """
    pass


def GetData(transaction_attr):
    """

    :param transaction_attr:
    """
    pass



from boa.blockchain.vm.Neo.Header import *
from boa.blockchain.vm.Neo.Block import *
from boa.blockchain.vm.Neo.Transaction import *
from boa.blockchain.vm.Neo.Account import *
from boa.blockchain.vm.Neo.Asset import *
from boa.blockchain.vm.Neo.Contract import *


def GetHeight() -> int:
    """

    """
    pass


def GetHeader(height_or_hash) -> Header:
    """

    :param height_or_hash:
    """
    pass


def GetBlock(height_or_hash) -> Block:
    """

    :param height_or_hash:
    """
    pass


def GetTransaction(hash) -> Transaction:
    """

    :param hash:
    """
    pass


def GetAccount(script_hash) -> Account:
    """

    :param script_hash:
    """
    pass


def GetValidators() -> []:
    """

    """
    pass


def GetAsset(asset_id) -> Asset:
    """

    :param asset_id:
    """
    pass


def GetContract(script_hash) -> Contract:
    """

    :param script_hash:
    """
    pass




def Main(a, b, c, d):

    """

    :param a:
    :param b:
    :param c:
    :param d:
    :return:
    """
    f = b * d

    g = d / c

    q = a + f

    m = q + g

    h = m - q

    j = h % g

    return j



from boa.code.builtins import range


def Main():

    """

    :return:
    """
    items = [0, 1, 2]

    items2 = [5, 6]
    items3 = [1, 2, 4, 5]
    count = 0

    q = 20
    for i in items:

        count += i

        if count < 20:

            for j in items2:
                count += j

                if q == 21:
                    break

                count = count + 1

                for blah in items3:
                    count = count + blah

    return count




from boa.code.builtins import concat


def Main(ba1, ba2):

    #    b = ba2

    # m = ba2[1] # subscribt for a byte array does not work

    """

    :param ba1:
    :param ba2:
    :return:
    """
    m = ba2[1:2]  # but you can do this instead

    # strings and byte arrays work the same
    mystr = 'staoheustnau'

    # this will not work
    # m = mystr[3]

    # but this will
    m = mystr[3:5]

    #
    m = ba1[1:]

    return concat(mystr, ba2)



from byteplay3 import Code, Opcode
from boa.code.method import Method
from boa.code import pyop
import importlib
import binascii


class Item():
    """

    """
    items = None

    def __init__(self, item_list):
        self.items = item_list

    def is_valid(self):
        """

        :return:
        """
        return True


class Definition(Item):
    pass
#    def __init__(self, item_list):
#        super(Definition, self).__init__(item_list)
#        self.items[-1] = (Opcode(pyop.STORE_FAST),self.items[-1][1])
#       print("self items %s " % self.items)


class Action(Item):

    """

    """
    event_name = None
    event_args = None

    method_name = None

    def __init__(self, item_list):
        super(Action, self).__init__(item_list)

        arguments = []

        for i, (key, value) in enumerate(self.items.items):
            if key == pyop.LOAD_CONST:
                arguments.append(value)
            elif key == pyop.STORE_NAME:
                self.method_name = value

        self.event_name = arguments[0]

        self.event_args = arguments


class SmartContractAppCall(Item):

    """

    """
    script_hash = None
    script_args = None

    method_name = None

    def __init__(self, item_list):
        super(SmartContractAppCall, self).__init__(item_list)

        arguments = []

        for i, (key, value) in enumerate(self.items.items):
            if key == pyop.LOAD_CONST:
                arguments.append(value)
            elif key == pyop.STORE_NAME:
                self.method_name = value

        self.script_hash = arguments[0]

        self.script_args = arguments

        if type(self.script_hash) is str:
            if len(self.script_hash) != 40:
                raise Exception(
                    "Invalid script hash! length of string must be 40")
        elif type(self.script_hash) in [bytes, bytearray]:
            if len(self.script_hash) != 20:
                raise Exception(
                    "Invalid Script hash, length in bytes must be 20")
        else:
            raise Exception(
                "Invalid script hash type.  must be string, bytes, or bytearray")

    @property
    def script_hash_addr(self):
        """

        :return:
        """

        return SmartContractAppCall.ToScriptHashData(self.script_hash)



    @staticmethod
    def ToScriptHashData(item):
        """

        :return:
        """
        b_array = None
        if type(item) is str:
            print("ITEM!! %s " % item)
            bstring = item.encode('utf-8')
            b_array = bytearray(binascii.unhexlify(bstring))
        elif type(item) is bytearray:
            pass
        elif type(item) is bytes:
            print("bytes len %s " % len(item))

            b_array = bytearray(item)
        else:
            raise Exception("Invalid script hash")

        b_array.reverse()

        return bytes(b_array)


class Import(Item):

    """

    """
    NEO_SC_FRAMEWORK = 'neo.SmartContract.Framework.'

    module_path = None
    module_name = None

    imported_module = None

    is_system_module = None

    module_items_to_import = None

    def __init__(self, item_list):
        super(Import, self).__init__(item_list)
        self.module_items_to_import = []

        for i, (op, arg) in enumerate(self.items):
            if op == pyop.IMPORT_NAME:
                self.module_path = arg
            elif op == pyop.STORE_NAME:
                self.module_items_to_import.append(arg)
#                print("SETTING MODALu nAme: %s " % self.module_name)

            elif op == pyop.IMPORT_STAR:
                self.module_items_to_import = ['STAR']

        self.is_system_module = False

        self.build()

    def build(self):
        # here is where we will check imports

        """

        """
        from boa.code.module import Module

        module = importlib.import_module(self.module_path, self.module_path)

        filename = module.__file__

        if self.NEO_SC_FRAMEWORK in self.module_path:
            self.is_system_module = True

        self.imported_module = Module(filename,
                                      module_name=self.module_path,
                                      is_sys_module=self.is_system_module,
                                      items_to_import=self.module_items_to_import)

    def is_valid(self):

        """

        :return:
        """
        return True

    def __str__(self):
        return "%s.%s" % (self.module_path, self.module_name)


class Klass(Item):

    """

    """
    name = None

    parent_name = None

    parent = None

    methods = None

    bp = None

    module = None

    def __init__(self, item_list, module):
        super(Klass, self).__init__(item_list)
        self.module = self.parent = module
        self.methods = []
        self.build()

    def build(self):

        """

        """
        for i, (op, arg) in enumerate(self.items):

            # if the item is a byteplay3 code object, it is a method
            if type(arg) is Code:
                self.bp = arg

            # load name is called  to gather the class parent
            if op == pyop.LOAD_NAME:
                self.parent_name = arg

            # this occurs to store the name of the class
            if op == pyop.STORE_NAME:
                self.name = arg

#        print('Created class %s inherits from %s ' % (self.name, self.parent))

        # go through code object of the class and extract the method code
        # objects
        for i, (op, arg) in enumerate(self.bp.code):

            if type(arg) is Code:
                self.methods.append(Method(arg, self))

    def is_valid(self):
        # here is where we check if the class extends something reasonable
        """

        :return:
        """
        return True




def Main():

    # lets have fun

    # this dosen't work. sad
    #        m = [[1,2,3],'fun','cool','neo']

    """

    :return:
    """
    m = [42, 'yay', 'neo', 'woolong', 'wtf-coin', b'\xff', 3, 10]

    answer = m[0]

    return answer




from byteplay3 import Opcode, Label
from boa.code.pytoken import PyToken
from boa.code import pyop
import pdb

class Block():

    """

    """
    forloop_counter = 0

    localmethod_counter = 0

    oplist = None  # list

    _label = None  # list

    iterable_variable = None
    iterable_loopcounter = None
    iterable_looplength = None
    iterable_item_name = None

    list_comp_iterable_variable = None
    list_comp_iterable_loopcounter = None
    list_comp_iterable_looplength = None
    list_comp_iterable_item_name = None

    slice_item_length = None

    has_dynamic_iterator = False

    local_func_name = None
    local_func_varname = None

    def __init__(self, operation_list):
        self.oplist = operation_list

        self.iterable_variable = None
        self.iterable_loopcounter = None
        self.iterable_looplength = None
        self.iterable_item_name = None

        self.list_comp_iterable_variable = None
        self.list_comp_iterable_loopcounter = None
        self.list_comp_iterable_looplength = None
        self.list_comp_iterable_item_name = None

        self.has_dynamic_iterator = False

        self.slice_item_length = None

    def __str__(self):
        if self._label:
            return '[Block] %s          [label] %s' % (self.oplist, self._label)
        return '[Block]: %s' % self.oplist

    def set_label(self, label):
        """

        :param label:
        """
        self._label = label
        self.oplist[0].jump_label = label

    @property
    def line(self):
        """

        :return:
        """
        if len(self.oplist):
            token = self.oplist[0]
            return token.line_no
        return None

    @property
    def has_load_attr(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.LOAD_ATTR:
                return True
        return False

    @property
    def has_make_function(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.MAKE_FUNCTION:
                return True
        return False

    @property
    def has_slice(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.BUILD_SLICE:
                return True

    @property
    def is_return(self):
        """

        :return:
        """
        if len(self.oplist):
            token = self.oplist[-1]
            if token.py_op == pyop.RETURN_VALUE:
                return True
        return False

    @property
    def is_iter(self):
        """

        :return:
        """
        has_get_iter = False
        for token in self.oplist:
            if token.py_op == pyop.GET_ITER:
                has_get_iter = True
            elif token.py_op == pyop.MAKE_FUNCTION:
                return False
        return has_get_iter

    @property
    def iterable_local_vars(self):
        """

        :return:
        """
        return [
            self.iterable_looplength,
            self.iterable_loopcounter,
            self.iterable_item_name,
        ]

    @property
    def list_comp_iterable_local_vars(self):
        """

        :return:
        """
        return [
            self.list_comp_iterable_looplength,
            self.list_comp_iterable_loopcounter,
            self.list_comp_iterable_item_name,
            self.list_comp_iterable_variable,
        ]

    @property
    def has_unprocessed_method_calls(self):
        """

        :return:
        """
#        if self.has_slice:
#            return False
#        if self.is_list_comprehension:
#            return False
        for token in self.oplist:
            if token.py_op == pyop.CALL_FUNCTION and not token.func_processed:
                return True
        return False

    @property
    def has_unprocessed_array_sub(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.STORE_SUBSCR and not token.array_processed:
                return True
        return False

    @property
    def has_unprocessed_array(self):
        """

        :return:
        """
        for token in self.oplist:
            if token.py_op == pyop.BUILD_LIST and not token.array_processed:
                return True
        return False

    @property
    def is_list_comprehension(self):
        """

        :return:
        """
        if self.has_make_function:
            for token in self.oplist:
                if token.py_op == pyop.GET_ITER:
                    return True
        if self.list_comp_iterable_variable:
            return True
        return False

    def preprocess_load_attr(self, method):
        """

        :param method:
        """
        while self.has_load_attr:

            index_to_rep = -1
            new_call = None

            for index, token in enumerate(self.oplist):

                if token.py_op == pyop.LOAD_ATTR:

                    what_to_load = 'Get%s' % token.args

                    call_func = PyToken(
                        Opcode(pyop.CALL_FUNCTION), lineno=self.line, index=-1, args=what_to_load)
                    call_func.func_processed = True
                    call_func.func_name = what_to_load
                    call_func.func_params = [self.oplist[index - 1]]

                    index_to_rep = index
                    new_call = call_func

            if index_to_rep > -1 and new_call is not None:
                self.oplist[index_to_rep] = new_call
                del self.oplist[index_to_rep - 1]

    def preprocess_make_function(self, method):
        """

        :param method:
        """
        code_obj = self.oplist[0].args
        code_obj_name = self.oplist[1].args
        self.local_func_name = "%s_%s" % (
            code_obj_name, Block.localmethod_counter)
        Block.localmethod_counter += 1

        from boa.code.method import Method

        m = Method(code_object=code_obj, parent=method.parent,
                   make_func_name=self.local_func_name)
        method.parent.add_method(m)

        self.local_func_varname = self.oplist[-1].args

    def preprocess_slice(self):
        """
        this method processes slices of strings or byte arrays such as item[1:3]
        """
        index_to_remove = -1

        for index, token in enumerate(self.oplist):
            if token.py_op == pyop.BUILD_SLICE:
                # first, we want to take out the BINARY_SUBSC op, since we wont need it
                index_to_remove = index + 1

                # now we want to check the second item, for example item[2:4], we need to check 4
                # if you do item[2:], in python, normally the end is inferred
                # but we get None.
                # in that case, we need to convert None into the length of the item being sliced
                end_op = self.oplist[index - 1]

                if end_op.args is None:
                    # 0xffffff
                    # if you have a list greater than that length, well you'll have other problems than this
                    max_len = 16777215
                    end_op.args = max_len


        if index_to_remove > -1:
            del self.oplist[index_to_remove]


    def preprocess_iter(self):

        # in a better world this would be done in a more efficient way
        # for now this is kept to be as understandable as possible

        """

        """
        loopsetup = self.oplist[0]
        loopsetup.args = None
        loopsetup.jump_label = None

        # first we need to create a loop counter variable
        self.iterable_loopcounter = 'forloop_counter_%s' % Block.forloop_counter

        # load the value 0
        loopcounter_start_ld_const = PyToken(
            op=Opcode(pyop.LOAD_CONST), lineno=loopsetup.line_no, index=-1, args=0)
        # now store the loop counter
        loopcounter_store_fast = PyToken(op=Opcode(
            pyop.STORE_FAST), lineno=loopsetup.line_no, index=-1, args=self.iterable_loopcounter)

        # this loads the list that is going to be iterated over ( LOAD_FAST )
        # this will be removed... its added into the call get length token function params
        # unless this is a dynamic iteration, like for x in range(x,y)
        dynamic_iterable_items = []

        iterable_load = self.oplist[1]

        self.iterable_item_name = iterable_load.args

        if iterable_load.py_op == pyop.CALL_FUNCTION:

            self.has_dynamic_iterator = True

            self.iterable_item_name = 'forloop_dynamic_range_%s' % Block.forloop_counter

            dynamic_iterator_store_fast = PyToken(op=Opcode(pyop.STORE_FAST), lineno=loopsetup.line_no, index=-1,
                                                  args=self.iterable_item_name)

            # if we're calling a method in this for i in, like for i in range(x,y) then we need
            # to call the function
            dynamic_iterable_items = [
                iterable_load, dynamic_iterator_store_fast]

        # Now we need to get the length of that list, and store that as a local variable

        call_get_length_token = PyToken(
            op=Opcode(pyop.CALL_FUNCTION), lineno=loopsetup.line_no, args=1)
        call_get_length_token.func_params = [iterable_load]
        call_get_length_token.func_name = 'len'

        # now we need a variable name to store the length of the array
        self.iterable_looplength = 'forloop_length_%s' % Block.forloop_counter

        # now store the variable which is the output of the len(items) call
        looplength_store_op = PyToken(op=Opcode(
            pyop.STORE_FAST), lineno=loopsetup.line_no, index=-1, args=self.iterable_looplength)

        get_iter = self.oplist[2]
        for_iter = self.oplist[3]

        store_iterable_name = self.oplist[4]

        # set the iterable variable name ( for example, i ) so that the loop body can use it
        self.iterable_variable = store_iterable_name.args

        ld_loopcounter = PyToken(op=Opcode(
            pyop.LOAD_FAST), lineno=loopsetup.line_no, index=-1, args=self.iterable_loopcounter)

        ld_loop_length = PyToken(op=Opcode(
            pyop.LOAD_FAST), lineno=loopsetup.line_no, index=-1, args=self.iterable_looplength)

        new__compare_op = PyToken(
            op=Opcode(pyop.COMPARE_OP), lineno=loopsetup.line_no, index=-1, args='<')
        new__popjump_op = PyToken(op=Opcode(
            pyop.POP_JUMP_IF_FALSE), lineno=loopsetup.line_no, index=-1, args=for_iter.args)

        for_iter.args = None

        self.oplist = [
            loopsetup,  # SETUP_LOOP

            get_iter,  # GET_ITER, keep this in for now


            # the following 4 ops set up the iterator

            loopcounter_start_ld_const,  # LOAD_CONST 0
            loopcounter_store_fast,  # STORE_FAST forloopcounter_X

            # dynamic load loop stuff would go here

            call_get_length_token,  # CALL_FUNCTION 1

            looplength_store_op,  # STORE_FAST forloop_length_X


            # these last 5 ops controls the operation of the loop

            for_iter,  # tihs is the jump target for the end of the loop execution block

            ld_loopcounter,  # load in the loop counter LOAD_FAST forloopcounter_X

            ld_loop_length,  # load in the loop length LOAD_FAST forloop_length_X

            new__compare_op,  # COMPARE_OP <, this will compare foorloop_counter_X < forloop_length_X

            new__popjump_op  # POP_JUMP_IF_FALSE jumps to the loop exit when counter == length
        ]

        if len(dynamic_iterable_items):
            self.oplist.insert(4, dynamic_iterable_items[0])
            self.oplist.insert(5, dynamic_iterable_items[1])

        Block.forloop_counter += 1

    def process_iter_body(self, setup_block):

        """

        :param setup_block:
        """
        first_op = self.oplist[0]

        #
        # the following loads the iterated item into the block
        #

        # load the iterable collection
        ld_load_iterable = PyToken(op=Opcode(
            pyop.LOAD_FAST), lineno=first_op.line_no, index=-1, args=setup_block.iterable_item_name)

        # load the counter var
        ld_counter = PyToken(op=Opcode(pyop.LOAD_FAST), lineno=first_op.line_no,
                             index=-1, args=setup_block.iterable_loopcounter)

        # binary subscript of the iterable collection
        ld_subscript = PyToken(op=Opcode(pyop.BINARY_SUBSCR),
                               lineno=first_op.line_no, index=-1)

        # now store the iterated item
        st_iterable = PyToken(op=Opcode(
            pyop.STORE_FAST), lineno=first_op.line_no, index=-1, args=setup_block.iterable_variable)

        #
        # the following load the forloop counter and increments it
        #

        # load the counter var
        ld_counter_2 = PyToken(op=Opcode(
            pyop.LOAD_FAST), lineno=first_op.line_no, index=-1, args=setup_block.iterable_loopcounter)
        # load the constant 1
        increment_const = PyToken(
            op=Opcode(pyop.LOAD_CONST), lineno=first_op.line_no, index=-1, args=1)
        # add it to the counter
        increment_add = PyToken(
            op=Opcode(pyop.INPLACE_ADD), lineno=first_op.line_no, index=-1)
        # and store it again
        increment_store = PyToken(op=Opcode(
            pyop.STORE_FAST), lineno=first_op.line_no, index=-1, args=setup_block.iterable_loopcounter)

        self.oplist = [
            ld_load_iterable, ld_counter, ld_subscript, st_iterable,

            ld_counter_2, increment_const, increment_add, increment_store

        ] + self.oplist

    def preprocess_method_calls(self, orig_method):

        """

        :param orig_method:
        """
        while self.has_unprocessed_method_calls:
            start_index_change = None
            end_index_change = None
            changed_items = None

            for index, token in enumerate(self.oplist):
                #                print("TOKEN::: %s %s " % (token.jump_label, token.args))

                if token.py_op == pyop.CALL_FUNCTION and not token.func_processed:

                    token.func_processed = True
                    param_count = token.args

                    # why would param count be 256 when calling w/ kwargs?
                    # when keyword args are sent, the param count is 256 * num paramms?
                    if param_count % 256 == 0:
                        param_count = 2 * int(param_count / 256)

                    params = self.oplist[index - param_count:index]

                    call_method_op = self.oplist[index - param_count - 1]

                    call_method_type = call_method_op.py_op
                    call_method_name = call_method_op.args

                    # we need to check if this is a method
                    # that is local to this block's method
                    for key, value in orig_method.local_methods.items():
                        if key == call_method_name:
                            call_method_name = value

                    token.func_name = call_method_name
                    token.func_type = call_method_type

                    # if this method is the target of a jump
                    # or if one of its parameters is the target of a jump
                    # we need to catch that and use that jump label
                    # otherwise bad things
                    if call_method_op.jump_label is not None:
                        if len(params) > 0:
                            params[0].jump_label = call_method_op.jump_label
                        else:
                            token.jump_label = call_method_op.jump_label

                    token.func_params = params

                    changed_items = [token]

                    start_index_change = index - param_count - 1
                    end_index_change = index

            if start_index_change is not None and end_index_change is not None:
                tstart = self.oplist[0:start_index_change]
                tend = self.oplist[end_index_change + 1:]
                self.oplist = tstart + changed_items + tend

    def preprocess_array_subs(self):
        """

        """
        while self.has_unprocessed_array_sub:
            start_index_change = None
            end_index_change = None
            changed_items = None

            for index, token in enumerate(self.oplist):
                if token.py_op == pyop.STORE_SUBSCR and not token.array_processed:
                    token.array_processed = True
                    start_index_change = index - 3
                    end_index_change = index

                    item_to_sub = self.oplist[index - 3].args
                    array_to_sub = self.oplist[index - 2].args
                    index_to_sub_at = self.oplist[index - 1].args
                    changed_items = []

                    # load the array to set the item into
                    ld_op = PyToken(Opcode(pyop.LOAD_FAST),
                                    token.line_no, args=array_to_sub)
                    changed_items.append(ld_op)

                    # create the setitem op
                    settoken = PyToken(Opcode(
                        pyop.SETITEM), token.line_no, args=index_to_sub_at, array_item=item_to_sub)
                    changed_items.append(settoken)

            if start_index_change is not None and end_index_change is not None:
                tstart = self.oplist[0:start_index_change]
                tend = self.oplist[end_index_change + 2:]
                self.oplist = tstart + changed_items + tend

    def preprocess_arrays(self):

        """

        """
        while self.has_unprocessed_array:

            blist_start_index = None
            blist_end_index = None
            array_items = []
            for index, token in enumerate(self.oplist):
                if token.py_op == pyop.BUILD_LIST and not token.array_processed and blist_start_index is None:

                    num_list_items = token.args
                    blist_start_index = index - num_list_items
                    blist_end_index = index
                    array_items = self.oplist[index -
                                              num_list_items:num_list_items]
                    array_items.reverse()

                    token.array_processed = True

            if blist_start_index is not None:
                self.oplist = self.oplist[0:blist_start_index] + \
                    array_items + self.oplist[blist_end_index:]

    def mark_as_end(self):

        """

        """
        tstart = self.oplist[:-1]
        tend = self.oplist[-1:]

        newitems = [PyToken(Opcode(pyop.NOP), self.line),
                    #                    PyToken(pyop.DROP_BODY, self.line),
                    PyToken(Opcode(pyop.FROMALTSTACK), self.line),
                    PyToken(Opcode(pyop.DROP), self.line)]

        self.oplist = tstart + newitems + tend

    def preprocess_list_comprehension(self, method):

        # i apologize for the following

        # grab the list comprehestion code object and make a method out of it
        # we will use it later
        """

        :param method:
        """
        code_obj = self.oplist[0].args

        # now get rid of the first 3 ops for now
        self.oplist = self.oplist[3:]

        # setup a loop for the list comp
        setup_loop = PyToken(op=Opcode(pyop.SETUP_LOOP),
                             lineno=self.line, index=-1)

        # first we need to create a loop counter variable
        self.list_comp_iterable_loopcounter = 'list_comp_loop_counter_%s' % Block.forloop_counter

        # load the value 0
        loopcounter_start_ld_const = PyToken(
            op=Opcode(pyop.LOAD_CONST), lineno=self.line, index=-1, args=0)
        # now store the loop counter
        loopcounter_store_fast = PyToken(op=Opcode(pyop.STORE_FAST), lineno=self.line, index=-1,
                                         args=self.list_comp_iterable_loopcounter)

        # this loads the list that is going to be iterated over ( LOAD_FAST )
        # this will be removed... its added into the call get length token function params
        # unless this is a dynamic iteration, like for x in range(x,y)

        iterable_load = self.oplist[0]

        self.list_comp_iterable_item_name = iterable_load.args

        # the following is in the case that we're doing something like for i in range(x,y)
        dynamic_iterable_items = []
        if iterable_load.py_op == pyop.CALL_FUNCTION:
            self.has_dynamic_iterator = True
            self.iterable_item_name = 'forloop_dynamic_range_%s' % Block.forloop_counter
            dynamic_iterator_store_fast = PyToken(op=Opcode(pyop.STORE_FAST), lineno=self.line, index=-1,
                                                  args=self.iterable_item_name)
            # if we're calling a method in this for i in, like for i in range(x,y) then we need
            # to call the function
            dynamic_iterable_items = [
                iterable_load, dynamic_iterator_store_fast]

        # Now we need to get the length of that list, and store that as a local variable
        call_get_length_token = PyToken(
            op=Opcode(pyop.CALL_FUNCTION), lineno=self.line, args=1)
        call_get_length_token.func_processed = True
        call_get_length_token.func_params = [iterable_load]
        call_get_length_token.func_name = 'len'

        # now we need a variable name to store the length of the array
        self.list_comp_iterable_looplength = 'list_comp_loop_length_%s' % Block.forloop_counter

        # now store the variable which is the output of the len(items) call
        looplength_store_op = PyToken(op=Opcode(pyop.STORE_FAST), lineno=self.line, index=-1,
                                      args=self.list_comp_iterable_looplength)

        get_iter = self.oplist[1]

        for_iter_label = Label()
        jmp_if_false_label = Label()

        for_iter = PyToken(op=Opcode(pyop.FOR_ITER),
                           lineno=self.line, index=-1)
        for_iter.jump_label = for_iter_label

        end_block = PyToken(op=Opcode(pyop.POP_BLOCK),
                            lineno=self.line, index=-1)
        end_block.jump_label = jmp_if_false_label

        jmp_abs_back = PyToken(op=Opcode(pyop.JUMP_ABSOLUTE),
                               lineno=self.line, index=-1, args=for_iter_label)

        self.list_comp_iterable_variable = 'list_comp_local_i_%s' % Block.forloop_counter

        ld_loopcounter = PyToken(op=Opcode(pyop.LOAD_FAST), lineno=self.line, index=-1,
                                 args=self.list_comp_iterable_loopcounter)

        ld_loop_length = PyToken(op=Opcode(pyop.LOAD_FAST), lineno=self.line, index=-1,
                                 args=self.list_comp_iterable_looplength)

        new__compare_op = PyToken(
            op=Opcode(pyop.COMPARE_OP), lineno=self.line, index=-1, args='<')
        new__popjump_op = PyToken(op=Opcode(pyop.POP_JUMP_IF_FALSE), lineno=self.line, index=-1,
                                  args=jmp_if_false_label)

        # ok now we do the loop block stuff here

        #
        # the following loads the iterated item into the block
        #

        # load the iterable collection
        ld_load_iterable = PyToken(op=Opcode(pyop.LOAD_FAST), lineno=self.line, index=-1,
                                   args=self.list_comp_iterable_item_name)

        # load the counter var
        ld_counter = PyToken(op=Opcode(pyop.LOAD_FAST), lineno=self.line, index=-1,
                             args=self.list_comp_iterable_loopcounter)

        # binary subscript of the iterable collection
        ld_subscript = PyToken(
            op=Opcode(pyop.BINARY_SUBSCR), lineno=self.line, index=-1)

        # now store the iterated item
        st_iterable = PyToken(op=Opcode(pyop.STORE_FAST), lineno=self.line, index=-1,
                              args=self.list_comp_iterable_variable)

        #
        # the following load the forloop counter and increments it
        #

        # load the counter var
        ld_counter_2 = PyToken(op=Opcode(pyop.LOAD_FAST), lineno=self.line, index=-1,
                               args=self.list_comp_iterable_loopcounter)
        # load the constant 1
        increment_const = PyToken(
            op=Opcode(pyop.LOAD_CONST), lineno=self.line, index=-1, args=1)
        # add it to the counter
        increment_add = PyToken(
            op=Opcode(pyop.INPLACE_ADD), lineno=self.line, index=-1)
        increment_add.func_processed = True
        # and store it again
        increment_store = PyToken(op=Opcode(pyop.STORE_FAST), lineno=self.line, index=-1,
                                  args=self.list_comp_iterable_loopcounter)

        # and now we call the function of the list-comprehension

        list_comp_call_func = PyToken(
            op=Opcode(pyop.CALL_FUNCTION), lineno=self.line, index=-1, args=1)
        list_comp_call_func.func_name = self.local_func_name
        list_comp_call_func.func_params = [self.list_comp_iterable_variable]

        self.oplist = [
            setup_loop,  # SETUP_LOOP

            get_iter,  # GET_ITER, keep this in for now


            # the following 4 ops set up the iterator

            loopcounter_start_ld_const,  # LOAD_CONST 0
            loopcounter_store_fast,  # STORE_FAST forloopcounter_X

            # dynamic load loop stuff would go here



            call_get_length_token,  # CALL_FUNCTION 1

            looplength_store_op,  # STORE_FAST forloop_length_X


            # these last 5 ops controls the operation of the loop

            for_iter,  # tihs is the jump target for the end of the loop execution block

            ld_loopcounter,  # load in the loop counter LOAD_FAST forloopcounter_X

            ld_loop_length,  # load in the loop length LOAD_FAST forloop_length_X

            new__compare_op,  # COMPARE_OP <, this will compare foorloop_counter_X < forloop_length_X

            new__popjump_op,  # POP_JUMP_IF_FALSE jumps to the loop exit when counter == length

            # the following are the loop body items
            ld_load_iterable,

            ld_counter,

            ld_subscript,

            st_iterable,

            ld_counter_2,

            increment_const,

            increment_add,  # this is a hack... when the list_comp_call_func is processed, it
            # takes out a few things from the block
            # so we add it in twice (blerg...) so it gets put back in
            increment_store,

            # put call method of the list comp here...


            # now pop back to for_iter
            jmp_abs_back,

            end_block,
        ]

#        from boa.code.method import Method
#        internal_method = Method(code_object=code_obj,
#                                 parent=method.parent,
#                                 make_func_name=self.local_func_name,
#                                 is_list_comp_internal=True,
#                                 list_comp_iterable_name=self.list_comp_iterable_variable)


#        internal_ops = internal_method.blocks[0].oplist
#        print("internal ops %s " % internal_ops)
#        print(internal_method.tokenizer.to_s())
#        self.oplist = self.oplist[:-2] + internal_ops + self.oplist[-2:]
#        if len(dynamic_iterable_items):
#            self.oplist.insert(4, dynamic_iterable_items[0])
#            self.oplist.insert(5, dynamic_iterable_items[1])

        Block.forloop_counter += 1

    def process_list_comp_internal(self, list_comp_item_name):
        # get rid of first op

        """

        :param list_comp_item_name:
        """
        print("OP 0 %s " % self.oplist[0].py_op)
        if self.oplist[0].py_op == pyop.STORE_FAST:

            argname = self.oplist[0].args

            self.oplist[0] = PyToken(
                op=Opcode(pyop.LOAD_FAST), lineno=self.line, index=0, args=list_comp_item_name)

            method_call = self.oplist[1]

            for op in method_call.func_params:
                print("op in method call:: %s %s " % (op, op.args))
                if op.args == argname:
                    print("switch argname...")
                    op.args = list_comp_item_name

            self.oplist = self.oplist[0:-3]

#        pdb.set_trace()





def GetTrigger():
    """

    """
    pass


def CheckWitness(hash_or_pubkey):
    """

    :param hash_or_pubkey:
    """
    pass


def Log(message):
    """

    :param message:
    """
    pass


def Notify(arg):
    """

    :param arg:
    """
    pass



from boa.code import pyop

from byteplay3 import Label, isopcode, haslocal, Code

from opcode import opname

from boa.blockchain.vm import VMOp


class PyToken():

    """

    """
    py_op = None

    args = None

    line_no = None

    addr = None

    tokenizer = None

    jump_label = None

    array_processed = False

    array_item = None

    # method calling things

    func_processed = False

    func_params = None

    func_name = None
    func_type = None

    script_hash_token = None

    @property
    def op_name(self):
        """

        :return:
        """
        if type(self.py_op) is int:
            return opname[self.py_op]
        elif type(self.py_op) is Label:
            return 'Label %s ' % self.py_op
        return self.py_op

    @property
    def is_op(self):
        """

        :return:
        """
        return isopcode(self.py_op)

    @property
    def is_local(self):
        """

        :return:
        """
        return haslocal(self.py_op)

    @property
    def arg_s(self):
        """

        :return:
        """
        if self.args:
            return str(self.args)
        return ''

    def __init__(self, op, lineno, index=None, args=None, array_item=None):

        self.py_op = op

        self.args = args

        self.line_no = lineno

        self.addr = index

        self.array_item = array_item

    def __str__(self):
        arg = ''

        if self.args:
            if type(self.args) is Label:
                arg = str(self.args)
            else:
                arg = self.args
            return '%s      %s   %s  --> %s ' % (self.line_no, self.addr, self.op_name, arg)
        return '%s      %s   %s' % (self.line_no, self.addr, self.op_name)

    def to_vm(self, tokenizer, prev_token=None):

        """

        :param tokenizer:
        :param prev_token:
        :return:
        """
        self.tokenizer = tokenizer
        token = None

        if self.is_op:

            op = self.py_op

            if op == pyop.NOP:
                token = tokenizer.convert1(VMOp.NOP, self)

            elif op == pyop.RETURN_VALUE:
                token = tokenizer.convert1(VMOp.RET, self)

            # control flow
            elif op == pyop.BR_S:
                token = tokenizer.convert1(VMOp.JMP, self, data=self.args)

            elif op == pyop.JUMP_FORWARD:
                token = tokenizer.convert1(VMOp.JMP, self, data=bytearray(2))

            elif op == pyop.JUMP_ABSOLUTE:
                token = tokenizer.convert1(VMOp.JMP, self, data=bytearray(2))

            elif op == pyop.POP_JUMP_IF_FALSE:
                token = tokenizer.convert1(
                    VMOp.JMPIFNOT, self, data=bytearray(2))

            elif op == pyop.POP_JUMP_IF_TRUE:
                token = tokenizer.convert1(VMOp.JMPIF, self, data=bytearray(2))

            # loops
            elif op == pyop.SETUP_LOOP:
                token = tokenizer.convert1(VMOp.NOP, self)

            elif op == pyop.BREAK_LOOP:
                token = tokenizer.convert1(VMOp.JMP, self, data=bytearray(2))

            elif op == pyop.FOR_ITER:
                token = tokenizer.convert1(VMOp.NOP, self)

#            elif op == pyop.GET_ITER:
#                token = tokenizer.convert1(VMOp.NOP, self)

            elif op == pyop.POP_BLOCK:
                token = tokenizer.convert1(VMOp.NOP, self)

            elif op == pyop.FROMALTSTACK:
                token = tokenizer.convert1(VMOp.FROMALTSTACK, self)
            elif op == pyop.DROP:
                token = tokenizer.convert1(VMOp.DROP, self)
            elif op == pyop.XSWAP:
                token = tokenizer.convert1(VMOp.XSWAP, self)
            elif op == pyop.ROLL:
                token = tokenizer.convert1(VMOp.ROLL, self)

            # loading constants ( ie 1, 2 etc)
            elif op == pyop.LOAD_CONST:

                if type(self.args) is int:
                    token = tokenizer.convert_push_integer(self.args, self)
                elif type(self.args) is str:
                    str_bytes = self.args.encode('utf-8')
                    self.args = str_bytes
                    token = tokenizer.convert_push_data(self.args, self)
                elif type(self.args) is bytes:
                    token = tokenizer.convert_push_data(self.args, self)
                elif type(self.args) is bytearray:
                    token = tokenizer.convert_push_data(bytes(self.args), self)
                elif type(self.args) is bool:
                    token = tokenizer.convert_push_integer(self.args)
                elif type(self.args) == type(None):
                    token = tokenizer.convert_push_data(bytearray(0))
                elif type(self.args) == Code:
                    pass
                else:

                    raise Exception("Could not load type %s for item %s " % (
                        type(self.args), self.args))

            # storing / loading local variables
            elif op in [pyop.STORE_FAST, pyop.STORE_NAME]:
                token = tokenizer.convert_store_local(self)

            elif op in [pyop.LOAD_FAST, pyop.LOAD_NAME, pyop.LOAD_GLOBAL]:
                token = tokenizer.convert_load_local(self)

            # unary ops

#            elif op == pyop.UNARY_INVERT:
#                token = tokenizer.convert1(VMOp.INVERT, self)

            elif op == pyop.UNARY_NEGATIVE:
                token = tokenizer.convert1(VMOp.NEGATE, self)

            elif op == pyop.UNARY_NOT:
                token = tokenizer.convert1(VMOp.NOT, self)

#            elif op == pyop.UNARY_POSITIVE:
                # hmmm
#                token = tokenizer.convert1(VMOp.ABS, self)
#                pass

            # math
            elif op in [pyop.BINARY_ADD, pyop.INPLACE_ADD]:

                # we can't tell by looking up the last token what type of item it was
                # will need to figure out a different way of concatting strings
                #                if prev_token and type(prev_token.args) is str:
                #                    token = tokenizer.convert1(VMOp.CAT, self)
                #                else:
                token = tokenizer.convert1(VMOp.ADD, self)

            elif op in [pyop.BINARY_SUBTRACT, pyop.INPLACE_SUBTRACT]:
                token = tokenizer.convert1(VMOp.SUB, self)

            elif op in [pyop.BINARY_MULTIPLY, pyop.INPLACE_MULTIPLY]:
                token = tokenizer.convert1(VMOp.MUL, self)

            elif op in [pyop.BINARY_FLOOR_DIVIDE, pyop.BINARY_TRUE_DIVIDE, pyop.INPLACE_FLOOR_DIVIDE, pyop.INPLACE_TRUE_DIVIDE]:
                token = tokenizer.convert1(VMOp.DIV, self)

            elif op in [pyop.BINARY_MODULO, pyop.INPLACE_MODULO]:
                token = tokenizer.convert1(VMOp.MOD, self)

            elif op == [pyop.BINARY_OR, pyop.INPLACE_OR]:
                token = tokenizer.convert1(VMOp.BOOLOR, self)

            elif op == [pyop.BINARY_AND, pyop.INPLACE_AND]:
                token = tokenizer.convert1(VMOp.BOOLAND, self)

            elif op == [pyop.BINARY_XOR, pyop.INPLACE_XOR]:
                token = tokenizer.convert1(VMOp.XOR, self)

            elif op in [pyop.BINARY_LSHIFT, pyop.INPLACE_LSHIFT]:
                token = tokenizer.convert1(VMOp.SHL, self)

            elif op in [pyop.BINARY_RSHIFT, pyop.INPLACE_RSHIFT]:
                token = tokenizer.convert1(VMOp.SHR, self)

            # compare

            elif op == pyop.COMPARE_OP:

                if self.args == '>':
                    token = tokenizer.convert1(VMOp.GT, self)
                elif self.args == '>=':
                    token = tokenizer.convert1(VMOp.GTE, self)
                elif self.args == '<':
                    token = tokenizer.convert1(VMOp.LT, self)
                elif self.args == '<=':
                    token = tokenizer.convert1(VMOp.LTE, self)
                elif self.args == '==':
                    token = tokenizer.convert1(VMOp.NUMEQUAL, self)
                elif self.args == 'is':
                    token = tokenizer.convert1(VMOp.EQUAL, self)
                elif self.args == '!=':
                    token = tokenizer.convert1(VMOp.NUMNOTEQUAL, self)

            # arrays
            elif op == pyop.BUILD_LIST:
                token = tokenizer.convert_new_array(VMOp.NEWARRAY, self)
            elif op == pyop.SETITEM:
                token = tokenizer.convert_set_element(self, self.args)
#                token = tokenizer.convert1(VMOp.SETITEM,self, data=self.args)
            elif op == pyop.STORE_SUBSCR:
                # this wont occur because this op is preprocessed into a SETITEM op
                pass
            elif op == pyop.BINARY_SUBSCR:
                token = tokenizer.convert1(VMOp.PICKITEM, self)

            elif op == pyop.BUILD_SLICE:
                token = tokenizer.convert_build_slice(self)

            # strings

            elif op == pyop.CALL_FUNCTION:

                token = tokenizer.convert_method_call(self)


#            else:
#                print("OP NOT CONVERTED %s " % op)

        return token



from boa.blockchain.vm.Neo.Transaction import Transaction


class Block():

    @property
    def TransactionCount(self):
        """

        :return:
        """
        return GetTransactionCount(self)

    @property
    def Transactions(self):
        """

        :return:
        """
        return GetTransactions(self)


def GetTransactionCount(block :Block) -> int:
    """

    returns the number of transactions in a block


    """
    pass


def GetTransactions(block: Block) -> list:
    """
    returns a list of transactions contained in a block


    """
    pass


def GetTransaction(block: Block, index:int) -> Transaction:
    """

    :param block: the block to get the transaction from
    :param index: the index of the transaction within the lock


    """

    pass





def Main():

    """

    :return:
    """

    def h(a, b):
        """

        :param a:
        :param b:
        :return:
        """
        return a + b

    k = h(1, 8)

    def q(a, b):

        # this will not work at the moment... not sure why you
        # would want to do this
        # l = h(3, 4)

        """

        :param a:
        :param b:
        :return:
        """
        return a - b

    t = q(10, 3)

    return k + mmm() + t


def mmm():

    """

    :return:
    """
    return 9





def Main():

    """

    :return:
    """
    a = None  # this gets coerced to 0

    b = 1

    if a is None:  # this evaluates to true ( which it is )
        b = 2

    c = a + b  # this evaluates to b + 0, so in this case 2

    # in python, the expected behavior for this would be to raise
    # a type error, since you can't add NoneType to an int
    # in the vm, its all ok i guess

    return c



from byteplay3 import SetLinenoType, Label, Opcode

from boa.code.pytoken import PyToken
from boa.code.vmtoken import VMTokenizer
from boa.code.block import Block
from boa.code import pyop

import dis

import collections


class Method():
    """
    The method is the main unit of functionality.  Any method can take 0 to many arguments and return 1 value

    Each method contains a variable amount of lines, or ``boa.code.block.Block`` objects, which represent
    discrete units of functionality.

    Each line in split into discrete ``boa.code.pytoken.PyToken`` objects, which are tokens which would be used
    by a python interpreter.

    After this, each line is turned into a block.  Once a block is complete, a variable amount of processing
    happens on the block to turn it into something that the Neo Virtual Machine will understand.

    Once we have a list of processed blocks, we can string them together and feed them to the VMTokenizer

    The VMTokenizer is responsible for turning PyToken objects into VMToken objects.

    When the method has been tokenized, each token then has an address within the method.  Once these addresses are complete,
    the ``convert_jumps`` method is called to tell each flow control operation where (which address) it will need to jump to.
    """
    
    bp = None

    parent = None

    tokens = None

    tokenizer = None

    local_stores = None

    start_line_no = None

    blocks = None

    method_address = None

    dynamic_iterator_count = 0

    local_methods = None

    __make_func_name = None


    @property
    def name(self):
        """
        Get the name of this method.

        :return: the name of this method
        :rtype: str
        """
        
        return self.bp.name


    @property
    def full_name(self):
        """
        Get the full name of this method ( with module path ).

        :return: module namespaced name of the method
        :rtype: str
        """

        if self.__make_func_name is None:
            if len(self.module.module_path):
                return '%s.%s' % (self.module.module_path, self.name)
            return self.name
        return self.__make_func_name


    @property
    def args(self):
        """
        Return a list of arguments in this method.

        :return: list of arguments for this method
        :rtype: list
        """
        
        return self.bp.args
        

    @property
    def code(self):
        """
        Return the ``byteplay3`` code object.

        :return: the ``byteplay3`` code object of this method
        :rtype: ``byteplay3.Code``
        """
        
        return self.bp.code


    @property
    def vm_tokens(self):
        """
        Returns a list of all vm tokens in this method.

        :return: a list of vm tokens in this method
        :rtype: list
        """
        
        return self.tokenizer.vm_tokens
        

    @property
    def firstlineno(self):
        """
        Get the starting line number of this method.

        :return: starting line number
        :rtype: int
        """

        return self.bp.firstlineno
        

    @property
    def total_lines(self):
        """
        Get the total number of lines ( aka blocks ) in this method.

        :return: total number of lines
        :rtype: int
        """

        count = 0
        for index, (op, arg) in enumerate(self.code):
            if type(op) is SetLinenoType:
                count += 1

        return count
        

    @property
    def total_module_variables(self):
        """
        Get the total number of local variables.

        :return: the number of variables in this module
        :rtype: int
        """
        
        return len(self.module.module_variables)
        

    @property
    def module(self):
        """
        Retrieves the module this method is a member of.

        :return: the module this method is a member of
        :rtype: ``boa.code.module.Module``
        """
        
        from boa.code.module import Module

        if type(self.parent) is Module:
            return self.parent
        elif type(self.parent.parent) is Module:
            return self.parent.parent
        elif type(self.parent.parent.parent) is Module:
            return self.parent.parent.parent
            
        return None
        

    def __init__(self, code_object, parent, make_func_name=None):

        self.bp = code_object

        self.parent = parent

        self.__make_func_name = make_func_name

        self.read_module_variables()

        self.read_initial_tokens()

        self.process_block_groups()

        self.tokenize()

        self.convert_jumps()


    def print(self):
        """
        This method prints the output of the method's ``byteplay3`` object 
        as it would be seen by a python interpreter.
        Compare this with the ``boa.code.method.Method.to_dis()`` output 
        and you will see subtle differences.

        sample output:

        >>> method.print()
              2            STORE_FAST           j
              12         1 LOAD_CONST           9
              14         4 LOAD_CONST           <byteplay3.Code object at 0x10cb5ec88>
                         5 LOAD_CONST           'Main.<locals>.q'
                         6 MAKE_FUNCTION        0
                         7 STORE_FAST           q
              22         9 LOAD_FAST            q
                        10 LOAD_FAST            j
                        11 CALL_FUNCTION        1
                        12 STORE_FAST           m
              24        14 LOAD_FAST            m
                        15 RETURN_VALUE
        """

        print(self.code)
        

    def to_dis(self):
        """
        This method prints the output of the method as it would be seen 
        by a python interpreter.
        compare this to the output of the ``boa.code.method.Method.print()`` and 
        you will see some subtle differences.

        >>> method.to_dis()
          3             STORE_FAST               0 (j)
         12           0 LOAD_CONST               1 (9)
         14           6 LOAD_CONST               2 (<code object q at 0x10cbbc810, 
                                                   file "./boa/tests/src/LambdaTest.py", line 14>)
                      9 LOAD_CONST               3 ('Main.<locals>.q')
                     12 MAKE_FUNCTION            0
                     15 STORE_FAST               1 (q)
         22          18 LOAD_FAST                1 (q)
                     21 LOAD_FAST                0 (j)
                     24 CALL_FUNCTION            1 (1 positional, 0 keyword pair)
                     27 STORE_FAST               2 (m)
         24          30 LOAD_FAST                2 (m)
                     33 RETURN_VALUE
        """

        out = self.bp.to_code()
        dis.dis(out)
        

    def read_module_variables(self):
        """
        Take all module ``global`` variables and gives this method access to them.
        """

        for definition in self.module.module_variables:

            items = definition.items

            self.bp.code = items + self.bp.code


    def read_initial_tokens(self):
        """
        Take the initial set of tokens from the ``byteplay3`` code object and turn them into blocks.
        """

        self.blocks = []

        self.local_methods = collections.OrderedDict()

        self.local_stores = collections.OrderedDict()

        current_line_no = None

        block_group = None

        self.tokenizer = VMTokenizer(self)

        current_label = None

        current_loop_token = None

        for i, (op, arg) in enumerate(self.code):

            # print("[%s] %s  ->  %s " % (i, op, arg))

            if type(op) is SetLinenoType:

                current_line_no = arg

                if self.start_line_no is None:
                    self.start_line_no = current_line_no

                if block_group is not None:

                    self.blocks.append(Block(block_group))

                block_group = []

            elif type(op) is Label:

                current_label = op

            else:

                if op in [pyop.STORE_FAST, pyop.STORE_NAME, pyop.STORE_GLOBAL] and arg not in self.local_stores.keys():
                    length = len(self.local_stores)
                    self.local_stores[arg] = length

                token = PyToken(op, current_line_no, i, arg)

                if op == pyop.SETUP_LOOP:
                    current_loop_token = token

                if op == pyop.BREAK_LOOP and current_loop_token is not None:
                    token.args = current_loop_token.args
                    current_loop_token = None

                if current_label is not None:
                    token.jump_label = current_label
                    current_label = None

                block_group.append(token)

        if len(block_group):
            self.blocks.append(Block(block_group))
            

    def process_block_groups(self):
        """
        Takes the current blocks ( similar to lines in a method ) and
        processes them so they can be tokenized properly.
        """

        iter_setup_block = None

        for index, block in enumerate(self.blocks):

            # if it is a return block
            # we need to insert a jmp at the start of the block
            # for the vm
            if block.is_return:

                # this jump needs to jump 3 bytes.  why? stay tuned to find out
                block_addr = b'\x03\x00'

                ret_token = PyToken(Opcode(pyop.BR_S),
                                    block.line, args=block_addr)
                ret_token.jump_label = block.oplist[0].jump_label
                block.oplist[0].jump_label = None
                block.oplist.insert(0, ret_token)
                block.mark_as_end()
#                length = len(self.local_stores)
#                self.local_stores[block.local_return_name] = length

            if block.has_load_attr:
                block.preprocess_load_attr(self)

            if block.is_list_comprehension:
                block.preprocess_list_comprehension(self)
                for localvar in block.list_comp_iterable_local_vars:
                    if localvar in self.local_stores.keys():
                        pass
                    else:
                        length = len(self.local_stores)
                        self.local_stores[localvar] = length

            if block.has_make_function:
                block.preprocess_make_function(self)
                self.local_methods[block.local_func_varname] = block.local_func_name


            if block.has_unprocessed_array:
                block.preprocess_arrays()

            if block.has_unprocessed_array_sub:
                block.preprocess_array_subs()

            if block.has_unprocessed_method_calls:
                block.preprocess_method_calls(self)

            if block.has_slice:
                block.preprocess_slice()
                if block.slice_item_length is not None:
                    length = len(self.local_stores)
                    self.local_stores[block.slice_item_length] = length

            if iter_setup_block is not None:
                block.process_iter_body(iter_setup_block)
                iter_setup_block = None

            if block.is_iter and not block.is_list_comprehension:
                block.preprocess_iter()
                for localvar in block.iterable_local_vars:

                    if localvar in self.local_stores.keys():
                        pass
                    else:
                        length = len(self.local_stores)
                        self.local_stores[localvar] = length
                iter_setup_block = block
                self.dynamic_iterator_count += 1

        alltokens = []

        for block in self.blocks:
            if block.has_make_function:
                if block.is_list_comprehension:
                    alltokens = alltokens + block.oplist
            else:
                alltokens = alltokens + block.oplist
        self.tokens = alltokens

        for index, token in enumerate(self.tokens):
            token.addr = index
            

    def tokenize(self):
        """
        Turn a set of ``boa.code.pytoken.PyToken`` objects into ``boa.code.vmtoken.VMToken`` objects.
        """

        self.tokenizer.update_method_begin_items()
        prevtoken = None
        for t in self.tokens:
            t.to_vm(self.tokenizer, prevtoken)
            prevtoken = t
            

    def convert_jumps(self):
        """
        Convert jumps that occur from flow control items 
        such as if, else, for loops, while loops, and breaks.
        """

        for key, vm_token in self.tokenizer.vm_tokens.items():

            if vm_token.pytoken and type(vm_token.pytoken.args) is Label:

                label = vm_token.pytoken.args

                for key2, vm_token_target in self.tokenizer.vm_tokens.items():

                    if vm_token_target.pytoken and vm_token_target.pytoken.jump_label is not None:

                        jump_to_label = vm_token_target.pytoken.jump_label

                        if jump_to_label == label:

                            difference = vm_token_target.addr - vm_token.addr

                            vm_token.data = difference.to_bytes(2, 'little', signed=True)


    def write(self):
        """
        Write the current state of the tokenizer to a byte string.

        :return: a byte string of the current tokenizer
        :rtype: bytes
        """

        out = self.tokenizer.to_b()
        return out





def Main():

    """

    :return:
    """
    a = [1, 2, 3, 4, 5]

    # this does not work
    # e = add(a[1], a[2], a[3], a[4], a[0])

    a2 = a[1]

    a3 = a[2]

    e = add(1, 2, 3, 4, a2)

    first_item = get_first_item(a)

    return e + first_item


def add(a, b, c, d, e):

    """

    :param a:
    :param b:
    :param c:
    :param d:
    :param e:
    :return:
    """
    result = a + b + c + d + e

    return result


def get_first_item(array_item):

    """

    :param array_item:
    :return:
    """
    return array_item[0]




from boa.code.builtins import range


def Main():

    """

    :return:
    """
    a = range(100, 120)

    # b = a[4] # this will fail, since the range list is only 4 elements long
    # ( 0, 1, 2, 3 )

    b = a[3]

#    q = add_items(4, 4)

    return b




def Main():

    # array sizes once they are created are immutable

    """

    :return:
    """
    m = [2, 4, 1, 5 + 12]

    # we cant do this.  once you create an array you cant extend it
    # m[5] = 8

    # we cant do this for now either
#        m.append(8)

    m[2] = 7 + 10

    m2 = [9, 10, 11, 12]

    # we can change items
    m2[0] = 4

    # we can access items in the array!
    q = m[1]

    # m3 = m + m2 // VM wont support concatenating arrays for now

    # returns 1 aka true
    return m2[0] == q




class Asset():

    @property
    def AssetId(self):
        """

        :return:
        """
        return GetAssetId(self)

    @property
    def AssetType(self):
        """

        :return:
        """
        return GetAssetType(self)

    @property
    def Amount(self):
        """

        :return:
        """
        return GetAmount(self)

    @property
    def Available(self):
        """

        :return:
        """
        return GetAvailable(self)

    @property
    def Precision(self):
        """

        :return:
        """
        return GetPrecision(self)

    @property
    def Owner(self):
        """

        :return:
        """
        return GetOwner(self)

    @property
    def Admin(self):
        """

        :return:
        """
        return GetAdmin(self)

    @property
    def Issuer(self):
        """

        :return:
        """
        return GetIssuer(self)


def GetAssetId(asset):
    """

    :param asset:
    """
    pass


def GetAssetType(asset):
    """

    :param asset:
    """
    pass


def GetAmount(asset):
    """

    :param asset:
    """
    pass


def GetAvailable(asset):
    """

    :param asset:
    """
    pass


def GetPrecision(asset):
    """

    :param asset:
    """
    pass


def GetOwner(asset):
    """

    :param asset:
    """
    pass


def GetAdmin(asset):
    """

    :param asset:
    """
    pass


def GetIssuer(asset):
    """

    :param asset:
    """
    pass


def Create(asset_type, name, amount, precision, owner, admin, issuer):
    """

    :param asset_type:
    :param name:
    :param amount:
    :param precision:
    :param owner:
    :param admin:
    :param issuer:
    """
    pass


def Renew(asset, years):
    """

    :param asset:
    :param years:
    """
    pass



from boa.blockchain.vm.Neo.Runtime import Notify,GetTrigger,CheckWitness
from boa.blockchain.vm.Neo.Action import RegisterAction
from boa.blockchain.vm.Neo.TriggerType import Application,Verification
from boa.blockchain.vm.Neo.Blockchain import GetTransaction
from boa.blockchain.vm.Neo.TransactionType import InvocationTransaction
from boa.blockchain.vm.Neo.Transaction import *

from boa.blockchain.vm.System.ExecutionEngine import GetScriptContainer,GetExecutingScriptHash,GetCallingScriptHash,GetEntryScriptHash

from boa.blockchain.vm.Neo.Output import GetScriptHash, GetValue, GetAssetId
from boa.blockchain.vm.Neo.Input import GetHash,GetIndex
from boa.blockchain.vm.Neo.Storage import GetContext, Get, Put, Delete

from boa.code.builtins import range,concat,list,print_var

OWNER= b'\xaf\x12\xa8h{\x14\x94\x8b\xc4\xa0\x08\x12\x8aU\nci[\xc1\xa5'

NEO_ASSET_ID = b'\x9b|\xff\xda\xa6t\xbe\xae\x0f\x93\x0e\xbe`\x85\xaf\x90\x93\xe5\xfeV\xb3J\\"\x0c\xcd\xcfn\xfc3o\xc5'


onDeposit = RegisterAction('deposit','account','amount')

onWithdrawRequestApproved = RegisterAction('withdrawApproved','account','vin_requests')


def Main(operation,args):

    trigger = GetTrigger()

    if trigger == b'\x00':

        print("doing verification")

        # always allow owner to withdraw
        # this should be changed
        # to not allow the owner to withdraw from
        # vins that are earmarked for others
        if CheckWitness(OWNER):
            return True
        else:

            # if it is not the owner, we need to verify a few things
            can_widthraw = CanWithdraw()

            return can_widthraw



    elif trigger == b'\x10':


        if operation == 'deposit':

            deposit = DepositNeo()
            return deposit

        elif operation == 'withdrawalRequest':

            withdrawable_txs = VerifyWithdrawalRequest(args)
            return withdrawable_txs

        elif operation == 'getPending':
            account = args[0]
            withdrawable_txs = GetPendingWithdrawal(account)
            return withdrawable_txs

        elif operation == 'clearPending':
            account = args[0]
            result = DeletePendingWithdrawal(account)
            return result

        elif operation == 'balanceOf':
            account = args[0]
            result = BalanceOf(account)
            return result

        else:

            return 'unknown operation'

    return False



def CanWithdraw():

    return False


def DepositNeo():

    # get reference to the tx the invocation is in
    tx = GetScriptContainer()

    references = tx.References

    if len(references) < 1:
        print("no neo attached")
        return False

    # we need to determine who sent the tx
    reference = references[0]
    sender = GetScriptHash(reference)

    # this will keep track of how much neo was deposited
    value = 0

    # this is the contract's address
    receiver = GetExecutingScriptHash()

    # go through all the outputs
    # and check if the receiver is the contract
    # if that is true, we add the value of the output to the
    # total sum that was deposited

    for output in tx.Outputs:
        shash = GetScriptHash(output)
        if shash == receiver:

            output_asset_id = GetAssetId(output)

            if output_asset_id == NEO_ASSET_ID:

                output_val = GetValue(output)
                value = value + output_val

    if value > 0:

        # now we know value was deposited from the sender

        # check the current balance of the sender
        # add it to the deposit amount, and save
        context = GetContext()
        current_balance = Get(context, sender)
        new_balance = current_balance + value
        Put(context, sender, new_balance)

        # send deposit event
        onDeposit(sender,value)

        return True

    return False


def VerifyWithdrawalRequest(args):

    account = args[0]

    current_balance = BalanceOf(account)

    if current_balance == 0:
        print("No Current balance")
        return False

    if CheckHasPendingWithdrawal(account):
        print("already a pending withdrawal")
        return False

    arglen = len(args)

    txids_len = arglen / 2


    vin_requests= list(length=txids_len)


    # so, this is a bit involved, so bear with me here
    # lets assume a person has a balance that the contract owes them
    # we need to allow them to put a hold on some tx vins that they can 'claim'
    # after they have put the hold on them, they can withdraw and in the
    # verification portion of the contract it will allow them

    # but they may need to put a hold on tx vins that have a value
    # more than they are owed.  So we want to allow them to put a hold
    # on x number of txids, where the amount contained in x-1 vins is less than the amount they are owed
    # and the amount in x txids is greater than or equal to the amount they are owed
    # but only by one


    hold_amount = 0
    okcount = 0


    for i in range(0, txids_len):

        j = i * 2

        request_vin_tx = args[j + 1]
        request_vin_index = args[j + 2]

        vin = LookupVIN(request_vin_tx, request_vin_index)

        if vin and vin[1] == NEO_ASSET_ID:

            actual_amount = vin[0]

            # if the amount being held is less than the current balance
            # it is ok, we allow the vin request
            # this allows the hold amount to go over
            # but if it is too much then invalidate
            # the entire request

            if hold_amount < current_balance:
                hold_amount += actual_amount

                vin_request = [request_vin_tx, request_vin_index]
                vin_requests[i] = vin_request
                okcount += 1

    if okcount != txids_len:
        print("Invalid txid request")
        return False

    Notify(hold_amount)
    Notify(current_balance)

    # now that we've checked the requestor isn't requesting too much
    # we will check that the requestor can put a hold on the vins
    # we make sure they can put a hold on all the vins
    # before actually placing the hold on the vin
    for vin in vin_requests:

        vin_tx = vin[0]
        vin_index = vin[1]

        has_hold = HasVINHold(vin_tx,vin_index)

        if has_hold:
            print("Cannot request withdrawal, vin(s) already have hold")
            return False


    # now that we're sure we can put holds on all the requested vins
    # we will put holds on the requested vins

    output = None

    for vin in vin_requests:
        vin_tx = vin[0]
        vin_index = vin[1]

        hold_req = PlaceVINHold(account, vin_tx, vin_index)

        # now we assemble a bytearray
        # to save to storage
        vin = concat(vin_index,vin_tx)

        if len(output) > 1:
            output = concat(output,vin)
        else:
            output = vin


    onWithdrawRequestApproved(account, vin_requests)

    m = SetPendingWithdrawal(account, output)

    return vin_requests


def BalanceOf(account):

    """
    Method to return the current balance of an address

    :param account: the account address to retrieve the balance for
    :type account: bytearray

    :return: the current balance of an address
    :rtype: int

    """
    context = GetContext()
    balance = Get(context, account)

    return balance


def CheckHasPendingWithdrawal(account):

    context = GetContext()

    account_has_pending= concat(account,'pending')

    pending = Get(context, account_has_pending)

    return pending


def SetPendingWithdrawal(account, value):

    context = GetContext()

    account_has_pending = concat(account, 'pending')

    if value == 0:

        Delete(context,account_has_pending)

    else:

        Put(context,account_has_pending,value)

    return True


def GetPendingWithdrawal(account):

    context = GetContext()

    account_pending = concat(account,'pending')

    result = Get(context, account_pending)

    return result


def DeletePendingWithdrawal(account):

    context = GetContext()
    account_pending = concat(account,'pending')

    Delete(context, account_pending)

    return True



def HasVINHold(txid, index):

    context = GetContext()
    hold_id = concat(txid, index)

    item = Get(context, hold_id)

    return item


def PlaceVINHold( account, txid, index):

    context = GetContext()

    hold_id= concat(index,txid)

    Put(context, hold_id, account)

    return True



def GetTXInputs():

    tx = GetScriptContainer()

    inputs = tx.Inputs
    inputlen = len(inputs)

    results = list(length=inputlen)

    count = 0

    for input in tx.Inputs:

        txid = GetHash(input)
        index = GetIndex(input)
        item = [txid,index]
        results[count] = item
        count += 1

    return results


def LookupVIN(txid, index):


    tx = GetTransaction(txid)

    if tx:

        outputs = tx.Outputs

        output_len = len(outputs)

        if index < output_len:

            output = outputs[index]

            assetType = GetAssetId(output)
            assetAmount = GetValue(output)
            toret = [assetAmount,assetType]
            return toret

    print("could not lookup vin. TX or output not found")
    return False



def Main():

    # array sizes once they are created are immutable

    # lets have fun with strings

    """

    :return:
    """
    m = ['awesome', 'fun', 'cool', 'neo']

    q = m[1]

    m2 = [1, 'wat', 'huzzah', 8]

    q2 = m[3]

    m3 = [b'\x00', b'\x20', b'\xff', b'\xa0']

    q3 = m3[3]

    # hmmm, something weird here.  almost works...
    # m4 = [bytearray(b'\x0f\xf0'), bytearray(b'\xff\xff')]

    # this doesnt work
#        q4 = m4[1]

    return q3



from boa.code.builtins import concat


def Main():

    """

    :return:
    """
    str1 = 'hello'
    str2 = 'world'

    str3 = concat(str1, str2)

    # this ends up being 'worldhello'
    # need to reverse the parameters...

    return str3




def Main():

    """

    :return:
    """
    a = 0
    b = 0

    c = 22

    while a < 9:

        a = a + 1

        if a == 7:
            break

        while b < 4:

            b = b + 1

            while c > 20:

                c = c - 5

    return a + b + c  # expect 28




def Main():

    print('hello!')

    a = b'\x03'

    b = 3

    if a == b:

        print("equal!!")

    else:

        print("not equal!")



    if sample_method_call() == sample_method_call_2():
        print("method equal")
    else:
        print("method not equal")

    if a is b:

        print("is!")

    else:

        print("not is!")


    return True



def sample_method_call():

    return 2


def sample_method_call_2():

    return 2



def Main():

    """

    :return:
    """
    a = 1

    b = 2

    c = stuff()

    d = stuff2()

    e = stuff8()

    f = blah()

    h = prevcall()

    return a + c + d + f + b + h


def stuff():

    """

    :return:
    """
    a = 4

    b = 2

    return a + b


def stuff2():

    """

    :return:
    """
    a = 8

    j = 10

    return j - a


def prevcall():

    """

    :return:
    """
    return stuff()


def stuff8():

    """

    :return:
    """
    q = 'hello'

    return q


def blah():

    """

    :return:
    """
    return 1 + 8



from boa.blockchain.vm.Neo.Runtime import Log
from boa.code.builtins import concat


def Main(operation, items):

    """

    :param operation:
    :param items:
    :return:
    """
    j = 10

    if operation == 'dostuff':

        j = 3

        if len(items) == 2:

            bytes1 = items[0]
            bytes2 = items[1]

            len1 = len(bytes1)
            len2 = len(bytes2)

            total = concat(bytes1, bytes2)

#            j = len1 + len2

            if total == 137707327489:
                Log("awesome!")

            else:
                Log("bad")

        else:

            j = 23

    elif operation == 'dont':

        j = 4

    return j




# the following are python opcodes taken from the `opcode` module
# these have been constantized for easier access
# these are the opcodes used by python


# not to be confused with opcodes from neo.VM.OpCode,
# which are the opcodes for the neo vm

POP_TOP = 1
ROT_TWO = 2
ROT_THREE = 3
DUP_TOP = 4
DUP_TOP_TWO = 5

NOP = 9
UNARY_POSITIVE = 10
UNARY_NEGATIVE = 11
UNARY_NOT = 12

UNARY_INVERT = 15

BINARY_MATRIX_MULTIPLY = 16
INPLACE_MATRIX_MULTIPLY = 17

BINARY_POWER = 19
BINARY_MULTIPLY = 20

BINARY_MODULO = 22
BINARY_ADD = 23
BINARY_SUBTRACT = 24
BINARY_SUBSCR = 25
BINARY_FLOOR_DIVIDE = 26
BINARY_TRUE_DIVIDE = 27
INPLACE_FLOOR_DIVIDE = 28
INPLACE_TRUE_DIVIDE = 29

GET_AITER = 50
GET_ANEXT = 51
BEFORE_ASYNC_WITH = 52

INPLACE_ADD = 55
INPLACE_SUBTRACT = 56
INPLACE_MULTIPLY = 57

INPLACE_MODULO = 59
STORE_SUBSCR = 60
DELETE_SUBSCR = 61
BINARY_LSHIFT = 62
BINARY_RSHIFT = 63
BINARY_AND = 64
BINARY_XOR = 65
BINARY_OR = 66
INPLACE_POWER = 67
GET_ITER = 68
GET_YIELD_FROM_ITER = 69

PRINT_EXPR = 70
LOAD_BUILD_CLASS = 71
YIELD_FROM = 72
GET_AWAITABLE = 73

INPLACE_LSHIFT = 75
INPLACE_RSHIFT = 76
INPLACE_AND = 77
INPLACE_XOR = 78
INPLACE_OR = 79
BREAK_LOOP = 80
WITH_CLEANUP_START = 81
WITH_CLEANUP_FINISH = 82

RETURN_VALUE = 83
IMPORT_STAR = 84

YIELD_VALUE = 86
POP_BLOCK = 87
END_FINALLY = 88
POP_EXCEPT = 89


HAVE_ARGUMENT = 90              # Opcodes from here have an argument:

STORE_NAME = 90       # Index in name list
DELETE_NAME = 91      # ""
UNPACK_SEQUENCE = 92   # Number of tuple items

FOR_ITER = 93  # jrel op

UNPACK_EX = 94
STORE_ATTR = 95       # Index in name list
DELETE_ATTR = 96      # ""
STORE_GLOBAL = 97     # ""
DELETE_GLOBAL = 98    # ""
LOAD_CONST = 100       # Index in const list

# hasconst.append(100

LOAD_NAME = 101       # Index in name list
BUILD_TUPLE = 102      # Number of tuple items
BUILD_LIST = 103       # Number of list items
BUILD_SET = 104        # Number of set items
BUILD_MAP = 105        # Number of dict entries (upto 255
LOAD_ATTR = 106       # Index in name list
COMPARE_OP = 107       # Comparison operator

IMPORT_NAME = 108     # Index in name list
IMPORT_FROM = 109     # Index in name list

JUMP_FORWARD = 110    # Number of bytes to skip
JUMP_IF_FALSE_OR_POP = 111  # Target byte offset from beginning of code
JUMP_IF_TRUE_OR_POP = 112  # "jabs op"
JUMP_ABSOLUTE = 113        # "jabs op"
POP_JUMP_IF_FALSE = 114    # "jabs op"
POP_JUMP_IF_TRUE = 115     # "jabs op"

LOAD_GLOBAL = 116     # Index in name list

CONTINUE_LOOP = 119   # Target address jrel
SETUP_LOOP = 120      # Distance to target address jrel
SETUP_EXCEPT = 121    # "jrel"
SETUP_FINALLY = 122   # "jrel"

LOAD_FAST = 124        # Local variable number
STORE_FAST = 125       # Local variable number
DELETE_FAST = 126      # Local variable number

RAISE_VARARGS = 130    # Number of raise arguments (1, 2, or 3
CALL_FUNCTION = 131    # #args + (#kwargs << 8


MAKE_FUNCTION = 132    # Number of args with default values
BUILD_SLICE = 133      # Number of items
MAKE_CLOSURE = 134
LOAD_CLOSURE = 135

LOAD_DEREF = 136

STORE_DEREF = 137

DELETE_DEREF = 138

CALL_FUNCTION_VAR = 140     # #args + (#kwargs << 8
CALL_FUNCTION_KW = 141      # #args + (#kwargs << 8
CALL_FUNCTION_VAR_KW = 142  # #args + (#kwargs << 8

SETUP_WITH = 143

LIST_APPEND = 145
SET_ADD = 146
MAP_ADD = 147

LOAD_CLASSDEREF = 148

SETUP_ASYNC_WITH = 154

EXTENDED_ARG = 144

BUILD_LIST_UNPACK = 149
BUILD_MAP_UNPACK = 150
BUILD_MAP_UNPACK_WITH_CALL = 151
BUILD_TUPLE_UNPACK = 152
BUILD_SET_UNPACK = 153


# boa custom ops

FROMALTSTACK = 241
DROP = 242

BR_S = 243

SETITEM = 244

LD_ELEMENT = 245

XSWAP = 246
ROLL = 247

DROP_BODY = 248



from boa.blockchain.vm.Neo.Runtime import Notify


def Main():

    """

    :return:
    """
    a = 4

    b = a << 2

    c = a >> 2

    print("shifted a!")

    Notify(b)

    Notify(c)

    m = 16

    m >>= 3

    q = 670

    q >>= 8

    Notify(m)

    if q == m:
        print("all ok!")

    else:
        print("not ok")

    return b




from boa.code.builtins import concat


def Main():

    """

    :return:
    """
    c = b'\x01\x04\xaf\x09'

    l = len(c)

    b = c[2:l]

    j = b'\x01\x02\x03\x04\x05\x06\x07'

    k = concat(c, j)

    m = k[3:6]

    return concat(m, b)



from boa.blockchain.vm.Neo.Runtime import Notify

def Main():


    m = bytearray(b'\x01\x02\x03\x04\x05\x06\x07\x08')

    # this is a test to see if slice notation without specifying an end
    # is functional
    s1 = m[4:]

    Notify(s1) # should be b'\x05\x06\x07\x08'

    q = 2

    a = 1

    # this is a test to see if a bunch of different stuff works
    s2 = m[a+q-1:get_slice_end()-1]

    Notify(s2) # expected result should be b'\x03\x04'

    return s1



def get_slice_end():

    return 5



class TransactionInput():

    @property
    def Hash(self):
        """

        :return:
        """
        return GetHash(self)

    @property
    def Index(self):
        """

        :return:
        """
        return GetIndex(self)


def GetHash(input):
    """

    :param input:
    """
    pass


def GetIndex(input):
    """

    :param input:
    """
    pass




class Transaction():

    @property
    def Hash(self):
        """

        :return:
        """
        return GetHash(self)

    @property
    def Type(self):
        """

        :return:
        """
        return GetType(self)

    @property
    def Attributes(self):
        """

        :return:
        """
        return GetAttributes(self)

    @property
    def Inputs(self):
        """

        :return:
        """
        return GetInputs(self)

    @property
    def Outputs(self):
        """

        :return:
        """
        return GetOutputs(self)

    @property
    def References(self):
        """

        :return:
        """
        return GetReferences(self)


def GetHash(transaction):
    """

    :param transaction:
    """
    pass


def GetType(transaction):
    """

    :param transaction:
    """
    pass


def GetAttributes(transaction):
    """

    :param transaction:
    """
    pass


def GetInputs(transaction):
    """

    :param transaction:
    """
    pass


def GetOutputs(transaction):
    """

    :param transaction:
    """
    pass


def GetReferences(transaction):
    """

    :param transaction:
    """
    pass




from boa.code.builtins import list, range
from boa.blockchain.vm.Neo.Runtime import Log, Notify


def Main():

    """

    :return:
    """
    print("holla?")  # using pythonic print(), this is tranlated to Neo.Runtime.Log
    start = 4
    stop = 9

    r = range(start, stop)

    # using built in Neo.Runtime.Log( this is the same as print(message) )
    Log("hellllllllloo")

    # using the Neo.Runtime.Notify ( this is for logging variables... )
    Notify(start)

    l = list(length=stop)

    l[3] = 17

    b = r[3]
    print("hullo")

    return b



from boa.blockchain.vm.Neo.Runtime import Notify,GetTrigger,CheckWitness
from boa.blockchain.vm.Neo.Action import RegisterAction
from boa.blockchain.vm.Neo.TriggerType import Application,Verification

from boa.blockchain.vm.Neo.TransactionType import InvocationTransaction
from boa.blockchain.vm.Neo.Transaction import *

from boa.blockchain.vm.System.ExecutionEngine import GetScriptContainer,GetExecutingScriptHash
from boa.blockchain.vm.Neo.TriggerType import Application, Verification
from boa.blockchain.vm.Neo.Output import GetScriptHash, GetValue, GetAssetId
from boa.blockchain.vm.Neo.Storage import GetContext, Get, Put, Delete

OWNER= b'\x13\xff4\xcc\x10\x1cVs\x7fe\xc3\xb3\xd2\xf9iTHESK'

NEO_ASSET_ID = b'\x9b|\xff\xda\xa6t\xbe\xae\x0f\x93\x0e\xbe`\x85\xaf\x90\x93\xe5\xfeV\xb3J\\"\x0c\xcd\xcfn\xfc3o\xc5'


onDeposit = RegisterAction('deposit','account','amount')
onWithdraw = RegisterAction('withdraw','account','amount')
onWithdrawReconciled = RegisterAction('withdraw_reconcile', 'account','amount')

def Main(args):

    print("Do Withdraw Test")

    trigger = GetTrigger()

    if trigger == Verification():

        print("doing verification")

        # check to see if the withdrawer is the owner

        is_owner = CheckWitness(OWNER)

        # always allow owner to withdraw
        if is_owner:
            return True

        else:

            # if it is not the owner, we need to verify a few things
            print("will test can withdraw")
            can_widthraw = CanWithdrawNeo()

            print("can withdraw neo?")
#            Notify(can_widthraw)

            return can_widthraw



    elif trigger == Application():
        print("doing application")

        # no matter what happens, we need to reconcile balances
        # so if, for example, someone has sucessfully made a withdrawl
        # the verification portion of this contract checked it all out
        # but the application part of the contract needs to sync the balance

        # we don't really care what operation is called
        # no matter what we will try to reconcile
        did_withdraw = ReconcileBalances()


        operation = args[0]

        if operation == 'deposit':

            if not did_withdraw:
                deposit = DepositNeo()

                return deposit
            else:

                print("cannot withdraw and deposit at the same time")
                return False

        elif operation == 'balanceOf':

            if len(args) != 2:
                return False

            account = args[1]
            getbalance = BalanceOf(account)

            return getbalance


    return False




def CanWithdrawNeo():

    print("[can withdraw]")

    tx = GetScriptContainer()

    type = GetType(tx)

    print("[can withdraw] got tx type...")

    if type == 209:
        print("[can withdraw] Is invocation!!")

        # this is the contract's address
        sender_addr = GetExecutingScriptHash()


        withdrawal_amount = 0

        receiver_addr = bytearray(0)

        # go through the outputs of the tx
        # and find ones that are neo
        # and also the receiver cannot be the contract ( sender_addr )
        for output in tx.Outputs:
            shash = GetScriptHash(output)

            output_asset_id = GetAssetId(output)

            if output_asset_id == NEO_ASSET_ID:

                output_val = GetValue(output)

                if shash != sender_addr:

                    print("[can withdraw] output is to receiver")
                    receiver_addr = shash


                    withdrawal_amount = withdrawal_amount + output_val

                    Notify(withdrawal_amount)
                    Notify(output)

                else:

                    print("[can withdraw] output is to contract sender addr, subtract from withdraw total")

                    withdrawal_amount = withdrawal_amount - output_val

                    Notify(output)
            else:

                print("[can withdraw] output is not neo")
                Notify(output)


        # we check recevier addr and amount

        if len(receiver_addr) > 0:

            print("[can withdraw] receiver addr is valid")
            Notify(receiver_addr)

            context = GetContext()

            current_balance = Get(context, receiver_addr)

            print("[can withdraw] current balance")
            Notify(current_balance)
            Notify(withdrawal_amount)

            if withdrawal_amount <= current_balance:

                print("[can withdraw] withdrawl amount il less than balance")

                onWithdraw(receiver_addr, withdrawal_amount)

                return True

            else:

                print("[can withdraw] not enough to witdraw")
        else:

            print("[can withdraw] receiver addr not set")
    else:

        print("[can withdraw] tx is not invocation tx. return false")

    return False



def DepositNeo():

    # get reference to the tx the invocation is in
    tx = GetScriptContainer()

    references = tx.References

    if len(references) < 1:
        print("no neo attached")
        return False

    # we need to determine who sent the tx
    reference = references[0]
    sender = GetScriptHash(reference)

    # this will keep track of how much neo was deposited
    value = 0

    output_asset_id = GetAssetId(reference)

    if output_asset_id != NEO_ASSET_ID:
        print("Must deposit NEO")
        return False

    # this is the contract's address
    receiver = GetExecutingScriptHash()

    # go through all the outputs
    # and check if the receiver is the contract
    # if that is true, we add the value of the output to the
    # total sum that was deposited

    for output in tx.Outputs:
        shash = GetScriptHash(output)
        if shash == receiver:
            output_val = GetValue(output)
            value = value + output_val

    if value > 0:

        print("neo was deposited")
        Notify(value)

        # now we know value was deposited from the sender

        context = GetContext()

        #check the current balance of the sender
        current_balance = Get(context, sender)
        print("current balance...")
        Notify(current_balance)

        new_balance = current_balance + value

        print("new balance..")
        Notify(new_balance)

        Put(context, sender, new_balance)

        print("deposited")

        onDeposit(sender,value)

        return True

    return False


def BalanceOf(account):

    """
    Method to return the current balance of an address

    :param account: the account address to retrieve the balance for
    :type account: bytearray

    :return: the current balance of an address
    :rtype: int

    """
    print("getting balance of...")
    context = GetContext()
    print("getting context...")
    balance = Get(context, account)
    print("got balance...")

    if len(balance) < 1:
        balance = 0

    return balance



def ReconcileBalances():

    print("[Reconcile balances]")
    # this is the contract's address
    sender_addr = GetExecutingScriptHash()

    tx = GetScriptContainer()

    withdrawal_amount = 0

    receiver_addr = bytearray(0)

    # go through the outputs of the tx
    # and find ones that are neo
    # and also the receiver cannot be the contract ( sender_addr )
    for output in tx.Outputs:
        shash = GetScriptHash(output)

        output_asset_id = GetAssetId(output)

        if output_asset_id == NEO_ASSET_ID:

            if shash != sender_addr:

                print("[reconcile] output is to receiver")
                receiver_addr = shash

                output_val = GetValue(output)

                withdrawal_amount = withdrawal_amount + output_val

                Notify(withdrawal_amount)
                Notify(output)

            else:

                print("[Reconcile balances] output is to contract sender addr, ignore")
                Notify(output)
        else:

            print("[Reconcile balances] output is not neo")
            Notify(output)

    # we check recevier addr and amount

    if len(receiver_addr) > 0:

        print("[Reconcile balances] receiver addr is valid")
        Notify(receiver_addr)

        context = GetContext()

        current_balance = Get(context, receiver_addr)

        print("[Reconcile balances] current balance")
        Notify(current_balance)
        Notify(withdrawal_amount)

        if withdrawal_amount <= current_balance:

            new_balance = current_balance - withdrawal_amount

            print("[Reconcile balances] new balance is...")
            Notify(current_balance)
            Notify(new_balance)

            Put(context, receiver_addr, new_balance)

            onWithdrawReconciled(receiver_addr, new_balance)
            print("[Reconcile balances] withdrawl amount il less than balance")
            return True

        else:

            print("[Reconcile balances] not enough to witdraw")
    else:

        print("[Reconcile balances] receiver addr not set")

    return False



from boa.blockchain.vm.Neo.Runtime import Notify, Log


def Main():

    """

    :return:
    """
    a = 3

    b = 2

    m = 0

    if a < b:
        print("lt")

    else:

        print("Gt")
        m = 4

    print("ok!!")

    return m




def RegisterAppCall(smart_contract_hash, *args):
    """

    :param smart_contract_hash:
    :param args:
    """
    pass


def AppCall(smart_contract_hash, *args):
    """

    :param smart_contract_hash:
    :param args:
    """
    pass




from boa.blockchain.vm.Neo.Action import RegisterAction


Transfer = RegisterAction('transfer', 'from', 'to', 'amount')

Refund = RegisterAction('refund', 'to', 'amount')


def Main():

    """

    :return:
    """
    a = 2

    b = 5

    c = a + b

    Transfer(a, b, c)

    to = 'me'
    amount = 52

    Refund(to, amount)

    return c





def Main():

    """

    :return:
    """
    a = 1

    c = 3

    a += c

    b = 10

    b -= a

    d = 2

    b *= d

    b /= c

    b %= 3

    f = b + 20

    # f |= 34 # this doesn't curretly work

    return f  # expect 21
